import ast_toolbox.mcts.AdaptiveStressTesting as AST
import ast_toolbox.mcts.AdaptiveStressTestingBlindValue as AST_BV
from ast_toolbox.algos.mcts import MCTS


class MCTSBV(MCTS):
    """Monte Carlo Tress Search (MCTS) with double progressive widening (DPW) [1]_ using Blind Value search from Couetoux et al.  [2]_.

    Parameters
    ----------
    M : int, optional
        The number of randon decisions generated for the action pool.
    kwargs :
        Keyword arguments passed to `ast_toolbox.algos.mcts.MCTS`.

    References
    ----------
    .. [1] Lee, Ritchie, et al. "Adaptive stress testing of airborne collision avoidance systems."
     2015 IEEE/AIAA 34th Digital Avionics Systems Conference (DASC). IEEE, 2015.
    .. [2] Couetoux, Adrien, Hassen Doghmen, and Olivier Teytaud. "Improving the exploration in upper confidence trees."
     International Conference on Learning and Intelligent Optimization. Springer, Berlin, Heidelberg, 2012.
    """

    def __init__(self,
                 M=10,
                 **kwargs):
        self.ec = kwargs['ec']
        self.M = M
        super(MCTSBV, self).__init__(**kwargs)

    def init(self):
        """Initiate AST internal parameters
        """
        ast_params = AST.ASTParams(self.max_path_length, self.log_interval, self.log_tabular, self.log_dir, self.n_itr)
        ast_params.ec = self.ec
        ast_params.M = self.M
        self.ast = AST_BV.AdaptiveStressTestBV(p=ast_params, env=self.env, top_paths=self.top_paths)



"""Original parallel sampler pool backend."""
import pickle
import signal

import numpy as np
from dowel import logger
from garage.experiment import deterministic
from garage.sampler.stateful_pool import SharedGlobal
from garage.sampler.stateful_pool import singleton_pool
from garage.sampler.utils import rollout


def _worker_init(g, id):
    """Initialize a worker.

    Parameters
    ----------
    g : :py:class:`garage.sampler.stateful_pool.SharedGlobal`
        SharedGlobal class from :py:mod:`garage.sampler.stateful_pool`.
    id : int
        Worker id.
    """
    if singleton_pool.n_parallel > 1:
        import os
        os.environ['CUDA_VISIBLE_DEVICES'] = ''
    g.worker_id = id


def initialize(n_parallel):
    """Initialize the worker pool.

    SIGINT is blocked for all processes created in parallel_sampler to avoid
    the creation of sleeping and zombie processes.

    If the user interrupts run_experiment, there's a chance some processes
    won't die due to a dead lock condition where one of the children in the
    parallel sampler exits without releasing a lock once after it catches
    SIGINT.

    Later the parent tries to acquire the same lock to proceed with his
    cleanup, but it remains sleeping waiting for the lock to be released.
    In the meantime, all the process in parallel sampler remain in the zombie
    state since the parent cannot proceed with their clean up.

    Parameters
    ----------
    n_parallel : int
        Number of workers to run in parallel.
    """

    try:
        signal.pthread_sigmask(signal.SIG_BLOCK, [signal.SIGINT])
        singleton_pool.initialize(n_parallel)
        singleton_pool.run_each(_worker_init,
                                [(id, )
                                 for id in range(singleton_pool.n_parallel)])
    finally:
        signal.pthread_sigmask(signal.SIG_UNBLOCK, [signal.SIGINT])


def _get_scoped_g(g, scope):
    if scope is None:
        return g
    if not hasattr(g, 'scopes'):
        g.scopes = dict()
    if scope not in g.scopes:
        g.scopes[scope] = SharedGlobal()
        g.scopes[scope].worker_id = g.worker_id
    return g.scopes[scope]


def _worker_populate_task(g, env, policy, scope=None):
    g = _get_scoped_g(g, scope)
    g.env = pickle.loads(env)
    g.policy = pickle.loads(policy)


def _worker_terminate_task(g, scope=None):
    g = _get_scoped_g(g, scope)
    if getattr(g, 'env', None):
        g.env.close()
        g.env = None
    if getattr(g, 'policy', None):
        g.policy.terminate()
        g.policy = None


def populate_task(env, policy, scope=None):
    """Set each worker's env and policy.

    Parameters
    ----------
    env : :py:class:`ast_toolbox.envs.ASTEnv`
        The environment.
    policy : :py:class:`garage.tf.policies.Policy`
        The policy.
    scope : str
        Scope for identifying the algorithm.
        Must be specified if running multiple algorithms
        simultaneously, each using different environments
        and policies.
    """
    logger.log('Populating workers...')
    if singleton_pool.n_parallel > 1:
        singleton_pool.run_each(
            _worker_populate_task,
            [(pickle.dumps(env), pickle.dumps(policy), scope)] *
            singleton_pool.n_parallel)
    else:
        # avoid unnecessary copying
        g = _get_scoped_g(singleton_pool.G, scope)
        g.env = env
        g.policy = policy
    logger.log('Populated')


def terminate_task(scope=None):
    """Close each worker's env and terminate each policy.

    Parameters
    ----------
    scope : str
        Scope for identifying the algorithm.
        Must be specified if running multiple algorithms
        simultaneously, each using different environments
        and policies.

    """
    singleton_pool.run_each(_worker_terminate_task,
                            [(scope, )] * singleton_pool.n_parallel)


def close():
    """Close the worker pool."""
    singleton_pool.close()


def _worker_set_seed(_, seed):
    logger.log('Setting seed to %d' % seed)
    deterministic.set_seed(seed)


def set_seed(seed):
    """Set the seed in each worker.

    Parameters
    ----------
    seed : int
        The random seed to be used by the worker.
    """
    singleton_pool.run_each(_worker_set_seed,
                            [(seed + i, )
                             for i in range(singleton_pool.n_parallel)])


def _worker_set_policy_params(g, params, scope=None):
    g = _get_scoped_g(g, scope)
    g.policy.set_param_values(params)


def _worker_set_env_params(g, params, scope=None):
    g = _get_scoped_g(g, scope)
    g.env.set_param_values(params)


def _worker_collect_one_path(g, max_path_length, scope=None):
    g = _get_scoped_g(g, scope)
    path = rollout(g.env, g.policy, max_path_length=max_path_length)
    return path, len(path['rewards'])


def sample_paths(policy_params,
                 max_samples,
                 max_path_length=np.inf,
                 env_params=None,
                 scope=None):
    """Sample paths from each worker.

    Parameters
    ----------
    policy_params :
        parameters for the policy. This will be updated on each worker process
    max_samples : int
        desired maximum number of samples to be collected. The
        actual number of collected samples might be greater since all trajectories
        will be rolled out either until termination or until max_path_length is
        reached
    max_path_length : int, optional
        horizon / maximum length of a single trajectory
    scope : str
        Scope for identifying the algorithm.
        Must be specified if running multiple algorithms
        simultaneously, each using different environments
        and policies.
    """
    singleton_pool.run_each(_worker_set_policy_params,
                            [(policy_params, scope)] *
                            singleton_pool.n_parallel)

    if env_params is not None:
        singleton_pool.run_each(_worker_set_env_params,
                                [(env_params, scope)] *
                                singleton_pool.n_parallel)

    return singleton_pool.run_collect(_worker_collect_one_path,
                                      threshold=max_samples,
                                      args=(max_path_length, scope),
                                      show_prog_bar=True)



import time

import numpy as np


class DPWParams:
    """Structure that stores the parameters for the MCTS with DPW.

    Parameters
    ----------
    d : int
        The maximum searching depth.
    gamma : float
        The discount factor.
    ec : float
        The weight for the exploration bonus.
    n : int
        The mximum number of iterations.
    k : float
        The constraint parameter used in DPW: |N(s,a)|<=kN(s)^alpha.
    alpha : float
        The constraint parameter used in DPW: |N(s,a)|<=kN(s)^alpha.
    clear_nodes : bool
        Whether to clear redundant nodes in tree.
        Set it to True for saving memoray. Set it to False to better tree plotting.
    """

    def __init__(self, d, gamma, ec, n, k, alpha, clear_nodes):
        self.d = d  # search depth
        self.gamma = gamma  # discount factor
        self.ec = ec  # exploration constant
        self.n = n  # number of iterations
        self.k = k  # dpw parameters
        self.alpha = alpha  # dpw parameters
        self.clear_nodes = clear_nodes


class DPWModel:
    """The model used in the tree search.

    Parameters
    ----------
    model : :py:class:`ast_toolbox.mcts.MDP.TransitionModel`
        The transition model.
    getAction : function
        getAction(s, tree) returns the action used in rollout.
    getNextAction : function
        getNextAction(s, tree) returns the action used in exploration.
    """

    def __init__(self, model, getAction, getNextAction):
        self.model = model
        self.getAction = getAction  # expert action used in rollout
        self.getNextAction = getNextAction  # exploration strategy


class StateActionStateNode:
    """The structure storing the transition state-action-state.
    """

    def __init__(self):
        self.n = 0  # UInt64
        self.r = 0.0  # Float64


class StateActionNode:
    """The structure representing the state-action node.
    """

    def __init__(self):
        self.s = {}  # Dict{State,StateActionStateNode}
        self.n = 0  # UInt64
        self.q = 0.0  # Float64


class StateNode:
    """The structure representing the state node.
    """

    def __init__(self):
        self.a = {}  # Dict{Action,StateActionNode}
        self.n = 0  # UInt64


class DPWTree:
    """The structure storing the seaching tree.
    """

    def __init__(self, p, f):
        self.s_tree = {}  # Dict{State,StateNode}
        self.p = p  # DPWParams
        self.f = f  # DPWModel


def saveBackwardState(old_s_tree, new_s_tree, s_current):
    """Saving the s_current as well as all its predecessors in the old_s_tree into the new_s_tree.

    Parameters
    ----------
    old_s_tree : dict
        The old tree.
    new_s_tree : dict
        The new tree.
    s_current : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The current state.

    Returns
    ----------
    new_s_tree : dict
        The new tree.
    """
    if not (s_current in old_s_tree):
        return new_s_tree
    s = s_current
    while s is not None:
        new_s_tree[s] = old_s_tree[s]
        s = s.parent
    return new_s_tree


def saveForwardState(old_s_tree, new_s_tree, s):
    """Saving the s_current as well as all its successors in the old_s_tree into the new_s_tree.

    Parameters
    ----------
    old_s_tree : dict
        The old tree.
    new_s_tree : dict
        The new tree.
    s_current : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The current state.

    Returns
    ----------
    new_s_tree : dict
        The new tree.
    """
    if not (s in old_s_tree):
        return new_s_tree
    new_s_tree[s] = old_s_tree[s]
    for sa in old_s_tree[s].a.values():
        for s1 in sa.s.keys():
            saveForwardState(old_s_tree, new_s_tree, s1)
    return new_s_tree


def saveState(old_s_tree, s):
    """Saving the s_current as well as all its predecessors and successors in the old_s_tree into the new_s_tree.

    Parameters
    ----------
    old_s_tree : dict
        The old tree.
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The current state.

    Returns
    ----------
    new_s_tree : dict
        The new tree.
    """
    new_s_tree = {}
    saveBackwardState(old_s_tree, new_s_tree, s)
    saveForwardState(old_s_tree, new_s_tree, s)
    return new_s_tree


def selectAction(tree, s, verbose=False):
    """Run MCTS to select one action for the state s

    Parameters
    ----------
    tree : :py:class:`ast_toolbox.mcts.MCTSdpw.DPWTree`
        The seach tree.
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The current state.
    verbose : bool, optional
        Where to log the seaching information.

    Returns
    ----------
    action : `ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`
        The selected AST action.
    """
    if tree.p.clear_nodes:
        new_dict = saveState(tree.s_tree, s)
        tree.s_tree.clear()
        tree.s_tree = new_dict

    depth = tree.p.d
    time.time() * 1e6
    for i in range(tree.p.n):
        R, actions = tree.f.model.goToState(s)
        R += simulate(tree, s, depth, verbose=verbose)

    tree.f.model.goToState(s)
    state_node = tree.s_tree[s]
    explored_actions = list(state_node.a.keys())
    nA = len(explored_actions)
    Q = np.zeros(nA)
    for i in range(nA):
        Q[i] = state_node.a[explored_actions[i]].q
    assert len(Q) != 0
    i = np.argmax(Q)
    return explored_actions[i]


def simulate(tree, s, depth, verbose=False):
    """Single run of the forward MCTS search.

    Parameters
    ----------
    tree : :py:class:`ast_toolbox.mcts.MCTSdpw.DPWTree`
        The seach tree.
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The current state.
    depth : int
        The maximum search depth
    verbose : bool, optional
        Where to log the seaching information.

    Returns
    ----------
    q : float
        The estimated return.
    """
    if (depth == 0) | tree.f.model.isEndState(s):
        return 0.0

    if not (s in tree.s_tree):
        tree.s_tree[s] = StateNode()
        return rollout(tree, s, depth)

    tree.s_tree[s].n += 1
    if len(tree.s_tree[s].a) < tree.p.k * tree.s_tree[s].n**tree.p.alpha:
        # explore new action
        a = tree.f.getNextAction(s, tree.s_tree)
        if not (a in tree.s_tree[s].a):
            tree.s_tree[s].a[a] = StateActionNode()
    else:
        # sample explored actions
        state_node = tree.s_tree[s]
        explored_actions = list(state_node.a.keys())
        nA = len(explored_actions)
        UCT = np.zeros(nA)
        nS = state_node.n
        assert nS > 0
        for i in range(nA):
            state_action_node = state_node.a[explored_actions[i]]
            assert state_action_node.n > 0
            UCT[i] = state_action_node.q + tree.p.ec * np.sqrt(np.log(nS) / float(state_action_node.n))
        a = explored_actions[np.argmax(UCT)]

    tree.s_tree[s].a[a].q

    sp, r = tree.f.model.getNextState(s, a)
    if not (sp in tree.s_tree[s].a[a].s):
        tree.s_tree[s].a[a].s[sp] = StateActionStateNode()
        tree.s_tree[s].a[a].s[sp].r = r
        tree.s_tree[s].a[a].s[sp].n = 1
    else:
        tree.s_tree[s].a[a].s[sp].n += 1

    q = r + tree.p.gamma * simulate(tree, sp, depth - 1)
    state_action_node = tree.s_tree[s].a[a]
    state_action_node.n += 1
    state_action_node.q += (q - state_action_node.q) / float(state_action_node.n)
    tree.s_tree[s].a[a] = state_action_node

    return q


def rollout(tree, s, depth):
    """Rollout from the current state s.

    Parameters
    ----------
    tree : :py:class:`ast_toolbox.mcts.MCTSdpw.DPWTree`
        The seach tree.
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The current state.
    depth : int
        The maximum search depth

    Returns
    ----------
    q : float
        The estimated return.
    """
    if (depth == 0) | tree.f.model.isEndState(s):
        return 0.0
    else:
        a = tree.f.getAction(s, tree.s_tree)
        sp, r = tree.f.model.getNextState(s, a)
        qval = (r + rollout(tree, sp, depth - 1))
        return qval



import pickle

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.models import LSTMModel
from tests.fixtures import TfGraphTestCase


class TestLSTMModel(TfGraphTestCase):
    def setup_method(self):
        super().setup_method()
        self.batch_size = 1
        self.time_step = 1
        self.feature_shape = 2
        self.output_dim = 1

        self.obs_inputs = np.full(
            (self.batch_size, self.time_step, self.feature_shape), 1.)
        self.obs_input = np.full((self.batch_size, self.feature_shape), 1.)

        self._input_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, None, self.feature_shape), name='input')
        self._step_input_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, self.feature_shape), name='input')

    @pytest.mark.parametrize('output_dim, hidden_dim', [(1, 1), (1, 2),
                                                        (3, 3)])
    def test_output_values(self, output_dim, hidden_dim):
        model = LSTMModel(
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            hidden_nonlinearity=None,
            recurrent_nonlinearity=None,
            hidden_w_init=tf.constant_initializer(1),
            recurrent_w_init=tf.constant_initializer(1),
            output_w_init=tf.constant_initializer(1))

        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_hidden',
            dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_cell',
            dtype=tf.float32)

        outputs = model.build(self._input_var, self._step_input_var,
                              step_hidden_var, step_cell_var)
        output = self.sess.run(
            outputs[0], feed_dict={self._input_var: self.obs_inputs})
        expected_output = np.full(
            [self.batch_size, self.time_step, output_dim], hidden_dim * 8)
        assert np.array_equal(output, expected_output)

    def test_is_pickleable(self):
        model = LSTMModel(output_dim=1, hidden_dim=1)
        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, 1), name='step_hidden', dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, 1), name='step_cell', dtype=tf.float32)
        model.build(self._input_var, self._step_input_var, step_hidden_var,
                    step_cell_var)

        # assign bias to all one
        with tf.compat.v1.variable_scope('LSTMModel/lstm', reuse=True):
            init_hidden = tf.compat.v1.get_variable('initial_hidden')

        init_hidden.load(tf.ones_like(init_hidden).eval())

        hidden = np.zeros((self.batch_size, 1))
        cell = np.zeros((self.batch_size, 1))

        outputs1 = self.sess.run(
            model.networks['default'].all_output,
            feed_dict={self._input_var: self.obs_inputs})
        output1 = self.sess.run(
            [
                model.networks['default'].step_output,
                model.networks['default'].step_hidden,
                model.networks['default'].step_cell
            ],
            feed_dict={
                self._step_input_var: self.obs_input,
                step_hidden_var: hidden,
                step_cell_var: cell
            })

        h = pickle.dumps(model)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            model_pickled = pickle.loads(h)

            input_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=(None, None, self.feature_shape),
                name='input')
            step_input_var = tf.compat.v1.placeholder(
                tf.float32, shape=(None, self.feature_shape), name='input')
            step_hidden_var = tf.compat.v1.placeholder(
                shape=(self.batch_size, 1),
                name='initial_hidden',
                dtype=tf.float32)
            step_cell_var = tf.compat.v1.placeholder(
                shape=(self.batch_size, 1),
                name='initial_cell',
                dtype=tf.float32)

            model_pickled.build(input_var, step_input_var, step_hidden_var,
                                step_cell_var)
            outputs2 = sess.run(
                model_pickled.networks['default'].all_output,
                feed_dict={input_var: self.obs_inputs})
            output2 = sess.run(
                [
                    model_pickled.networks['default'].step_output,
                    model_pickled.networks['default'].step_hidden,
                    model_pickled.networks['default'].step_cell
                ],
                feed_dict={
                    step_input_var: self.obs_input,
                    step_hidden_var: hidden,
                    step_cell_var: cell
                })
            assert np.array_equal(outputs1, outputs2)
            assert np.array_equal(output1, output2)



import csv

from matplotlib import pyplot as plt

csv_path = './Data/AST/Oct21/GAInter/0/process.csv'

with open(csv_path) as csv_file:
    csv_reader = csv.reader(csv_file, delimiter=',')
    rewards = []
    iterations = []
    for (i, row) in enumerate(csv_reader):
        print(i)
        if i == 0:
            entry_dict = {}
            for index in range(len(row)):
                entry_dict[row[index]] = index
        else:
            rewards.append(max(0.0, float(row[entry_dict["MaxReturn"]])))
            iterations.append(int(row[entry_dict["Iteration"]]))

fig = plt.figure()
plt.scatter(iterations, rewards)
plt.xlabel('Iteration')
plt.ylabel('MaxReturn')
fig.savefig('Data/Plot/' + 'cartpole_gaInterPopulationEvolve.pdf')
plt.close(fig)



"""MLP Merge Model.

A model composed only of a multi-layer perceptron (MLP), which maps
real-valued inputs to real-valued outputs. This model is called an
MLP Merge Model because it takes two inputs and concatenates the second
input with the layer at a specified index. It can be merged with any layer
from the input layer to the last hidden layer.
"""
import tensorflow as tf

from garage.tf.models.base import Model
from garage.tf.models.mlp import mlp


class MLPMergeModel(Model):
    """MLP Merge Model.

    Args:
        output_dim (int): Dimension of the network output.
        name (str): Model name, also the variable scope.
        hidden_sizes (list[int]): Output dimension of dense layer(s).
            For example, (32, 32) means this MLP consists of two
            hidden layers, each with 32 hidden units.
        concat_layer (int): The index of layers at which to concatenate
            input_var2 with the network. The indexing works like standard
            python list indexing. Index of 0 refers to the input layer
            (input_var1) while an index of -1 points to the last hidden
            layer. Default parameter points to second layer from the end.
        hidden_nonlinearity (callable): Activation function for intermediate
            dense layer(s). It should return a tf.Tensor. Set it to
            None to maintain a linear activation.
        hidden_w_init (callable): Initializer function for the weight
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        hidden_b_init (callable): Initializer function for the bias
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        output_nonlinearity (callable): Activation function for output dense
            layer. It should return a tf.Tensor. Set it to None to
            maintain a linear activation.
        output_w_init (callable): Initializer function for the weight
            of output dense layer(s). The function should return a
            tf.Tensor.
        output_b_init (callable): Initializer function for the bias
            of output dense layer(s). The function should return a
            tf.Tensor.
        layer_normalization (bool): Bool for using layer normalization or not.
    """

    def __init__(self,
                 output_dim,
                 name='MLPMergeModel',
                 hidden_sizes=(32, 32),
                 concat_layer=-2,
                 hidden_nonlinearity=tf.nn.relu,
                 hidden_w_init=tf.glorot_uniform_initializer(),
                 hidden_b_init=tf.zeros_initializer(),
                 output_nonlinearity=None,
                 output_w_init=tf.glorot_uniform_initializer(),
                 output_b_init=tf.zeros_initializer(),
                 layer_normalization=False):
        super().__init__(name)
        self._output_dim = output_dim
        self._hidden_sizes = hidden_sizes
        self._concat_layer = concat_layer
        self._hidden_nonlinearity = hidden_nonlinearity
        self._hidden_w_init = hidden_w_init
        self._hidden_b_init = hidden_b_init
        self._output_nonlinearity = output_nonlinearity
        self._output_w_init = output_w_init
        self._output_b_init = output_b_init
        self._layer_normalization = layer_normalization

    def network_input_spec(self):
        """Network input spec."""
        return ['input_var1', 'input_var2']

    def _build(self, state_input, action_input, name=None):
        return mlp(input_var=state_input,
                   output_dim=self._output_dim,
                   hidden_sizes=self._hidden_sizes,
                   input_var2=action_input,
                   concat_layer=self._concat_layer,
                   name='mlp_concat',
                   hidden_nonlinearity=self._hidden_nonlinearity,
                   hidden_w_init=self._hidden_w_init,
                   hidden_b_init=self._hidden_b_init,
                   output_nonlinearity=self._output_nonlinearity,
                   output_w_init=self._output_w_init,
                   output_b_init=self._output_b_init,
                   layer_normalization=self._layer_normalization)



"""AST-Toolbox Base"""
# from gym.envs.registration import register
import gym

__version__ = '2020.09.01.rc1'


def register(id, entry_point, force=True, **kwargs):
    env_specs = gym.envs.registry.env_specs
    if id in env_specs.keys():
        if not force:
            return
        del env_specs[id]
    gym.register(
        id=id,
        entry_point=entry_point,
        **kwargs
    )
# logger = logging.getLogger(__name__)


register(
    id='GoExploreAST-v0',
    entry_point='ast_toolbox.envs:GoExploreASTEnv',
    nondeterministic=True,
)

register(
    id='GoExploreAST-v1',
    entry_point='ast_toolbox.envs:Custom_GoExploreASTEnv',
    nondeterministic=True,
)
# register(
#     id='SoccerEmptyGoal-v0',
#     entry_point='gym_soccer.envs:SoccerEmptyGoalEnv',
#     timestep_limit=1000,
#     reward_threshold=10.0,
#     nondeterministic = True,
# )
#
# register(
#     id='SoccerAgainstKeeper-v0',
#     entry_point='gym.envs:SoccerAgainstKeeperEnv',
#     timestep_limit=1000,
#     reward_threshold=8.0,
#     nondeterministic = True,
# )



# Import the example classes
import pickle
import shelve

import fire
import gym
import numpy as np
# Useful imports
# from example_save_trials import *
import tensorflow as tf
from bsddb3 import db
from garage.envs.normalized_env import normalize
from garage.experiment import run_experiment
from garage.np.baselines.linear_feature_baseline import LinearFeatureBaseline
from garage.tf.envs.base import TfEnv
from garage.tf.experiment import LocalTFRunner

# Import the necessary garage classes
from ast_toolbox.algos import GoExplore
from ast_toolbox.policies import GoExplorePolicy
from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.samplers import BatchSampler
from ast_toolbox.simulators import ExampleAVSimulator
from ast_toolbox.spaces import ExampleAVSpaces

# Import the AST classes


#
# parser = argparse.ArgumentParser()
# parser.add_argument('--snapshot_mode', type=str, default="gap")
# parser.add_argument('--snapshot_gap', type=int, default=10)
# parser.add_argument('--log_dir', type=str, default='../data/')
# parser.add_argument('--iters', type=int, default=1)
# args = parser.parse_args()
#
# log_dir = args.log_dir
#
# batch_size = 4000
# max_path_length = 50
# n_envs = batch_size // max_path_length

# Example run command:
# python examples/AV/example_runner_ge_av.py runner --n_itr=1

def runner(
        env_args=None,
        run_experiment_args=None,
        sim_args=None,
        reward_args=None,
        spaces_args=None,
        policy_args=None,
        baseline_args=None,
        algo_args=None,
        runner_args=None,
        sampler_args=None,
        # log_dir='.',
):
    # exp_name='av',
    # use_ram=False,
    # db_filename='cellpool-shelf',
    # s_0=[0.0, -4.0, 1.0, 11.17, -35.0],
    # max_db_size=150,
    # overwrite_db=True,
    # n_parallel=1,
    # snapshot_mode='last',
    # snapshot_gap=1,
    #
    # max_path_length=50,
    # discount=0.99,
    # n_itr=100,
    # n_itr_robust=100,
    # max_kl_step=0.01,
    # whole_paths=False,
    # batch_size=None,
    # batch_size_robust=None):

    if env_args is dict and 'id' not in env_args.keys():
        print('ERROR: Must supply an environment id in env_args')
        raise Exception

    if run_experiment_args is None:
        run_experiment_args = {}

    if sim_args is None:
        sim_args = {}

    if reward_args is None:
        reward_args = {}

    if spaces_args is None:
        spaces_args = {}

    if policy_args is None:
        policy_args = {}

    if baseline_args is None:
        baseline_args = {}

    if algo_args is None:
        algo_args = {}

    if runner_args is None:
        runner_args = {'n_epochs': 1}

    if sampler_args is None:
        sampler_args = {}

    if 'n_parallel' in run_experiment_args:
        n_parallel = run_experiment_args['n_parallel']
    else:
        n_parallel = 1
        run_experiment_args['n_parallel'] = n_parallel

    if 'max_path_length' in sim_args:
        max_path_length = sim_args['max_path_length']
    else:
        max_path_length = 50
        sim_args['max_path_length'] = max_path_length

    if 'batch_size' in runner_args:
        batch_size = runner_args['batch_size']
    else:
        batch_size = max_path_length * n_parallel
        runner_args['batch_size'] = batch_size

        # if batch_size is None:
        #     batch_size = max_path_length * n_parallel
        #
        # if batch_size_robust is None:
        #     batch_size = max_path_length * n_parallel

    def run_task(snapshot_config, *_):

        config = tf.ConfigProto(device_count={'GPU': 0})
        # config.gpu_options.allow_growth = True
        with tf.Session(config=config) as sess:
            with tf.variable_scope('AST', reuse=tf.AUTO_REUSE):
                # Instantiate the example classes
                sim = ExampleAVSimulator(**sim_args)
                # blackbox_sim_state=True,
                # open_loop=False,
                # fixed_initial_state=True,
                # max_path_length=max_path_length)
                reward_function = ExampleAVReward(**reward_args)
                spaces = ExampleAVSpaces(**spaces_args)

                # Create the environment
                # env1 = GoExploreASTEnv(open_loop=False,
                #                              blackbox_sim_state=True,
                #                              fixed_init_state=True,
                #                              s_0=[-0.5, -4.0, 1.0, 11.17, -35.0],
                #                              simulator=sim,
                #                              reward_function=reward_function,
                #                              spaces=spaces
                #                              )
                # env1 = gym.make('ast_toolbox:GoExploreAST-v1',
                #                 blackbox_sim_state=True,
                #                 open_loop=False,
                #                 fixed_init_state=True,
                #                 s_0=s_0,
                #                 simulator=sim,
                #                 reward_function=reward_function,
                #                 spaces=spaces
                #                 )
                env1 = gym.make(id=env_args.pop('id'),
                                simulator=sim,
                                reward_function=reward_function,
                                spaces=spaces,
                                **env_args)
                env2 = normalize(env1)
                env = TfEnv(env2)

                # Instantiate the garage objects
                policy = GoExplorePolicy(env_spec=env.spec)

                baseline = LinearFeatureBaseline(env_spec=env.spec, **baseline_args)

                algo = GoExplore(env_spec=env.spec,
                                 env=env,
                                 policy=policy,
                                 baseline=baseline,
                                 **algo_args)
                #     db_filename=db_filename,
                #     max_db_size=max_db_size,
                #     env=env,
                #
                #     policy=policy,
                #     baseline=baseline,
                #     # robust_policy=robust_policy,
                #     # robust_baseline=robust_baseline,
                #     max_path_length=max_path_length,
                #     discount=discount,
                #     save_paths_gap=1,
                #     save_paths_path=log_dir,
                #     # whole_paths=whole_paths
                # )

                sampler_cls = BatchSampler
                # sampler_args = {'n_envs': n_parallel}
                sampler_args['sim'] = sim
                sampler_args['reward_function'] = reward_function

                with LocalTFRunner(snapshot_config=snapshot_config, sess=sess) as local_runner:
                    local_runner.setup(algo=algo,
                                       env=env,
                                       sampler_cls=sampler_cls,
                                       sampler_args=sampler_args)

                    # local_runner.setup(
                    #     algo=algo,
                    #     env=env,
                    #     sampler_cls=sampler_cls,
                    #     sampler_args={"sim": sim,
                    #                   "reward_function": reward_function})

                    # Run the experiment
                    best_cell = local_runner.train(**runner_args)  # n_epochs=n_itr, batch_size=batch_size, plot=False)

                    log_dir = run_experiment_args['log_dir']
                    db_filename = algo_args['db_filename']
                    s_0 = env_args['s_0']

                    pool_DB = db.DB()
                    pool_DB.open(db_filename + '_pool.dat', dbname=None, dbtype=db.DB_HASH, flags=db.DB_CREATE)
                    d_pool = shelve.Shelf(pool_DB, protocol=pickle.HIGHEST_PROTOCOL)
                    # pdb.set_trace()
                    print(best_cell)
                    temp = best_cell
                    paths = []
                    while (temp.parent is not None):
                        print(temp.observation)
                        action = temp.observation[1:].astype(np.float32) / 1000
                        paths.append({'state': temp.state,
                                      'reward': temp.reward,
                                      'action': action,
                                      'observation': np.array(s_0)})
                        temp = d_pool[temp.parent]
                    print(temp.observation)
                    paths.append({'state': temp.state,
                                  'reward': temp.reward,
                                  'action': action,
                                  'observation': np.array(s_0)})
                    # pdb.set_trace()
                    d_pool.close()

                    with open(log_dir + '/expert_trajectory.p', 'wb') as f:
                        pickle.dump([paths], f)
                    print('done!')

                    # Run backwards algorithm to robustify
                    # if n_itr_robust > 0:
                    #     with LocalRunner(
                    #             snapshot_config=snapshot_config, sess=sess) as local_runner:
                    #
                    #
                    #         robust_policy = GaussianLSTMPolicy(name='lstm_policy',
                    #                                            env_spec=env.spec,
                    #                                            hidden_dim=64,
                    #                                            use_peepholes=True)
                    #
                    #         robust_baseline = LinearFeatureBaseline(env_spec=env.spec)
                    #
                    #         optimizer = ConjugateGradientOptimizer
                    #         optimizer_args = {'hvp_approach': FiniteDifferenceHvp(base_eps=1e-5)}
                    #
                    #         robust_algo = BackwardAlgorithm(
                    #                             env=env,
                    #                             env_spec=env.spec,
                    #                             policy=robust_policy,
                    #                             baseline=robust_baseline,
                    #                             expert_trajectory=paths,
                    #                             epochs_per_step = 10,
                    #                             # scope=None,
                    #                             max_path_length=max_path_length,
                    #                             discount=discount,
                    #                             # gae_lambda=1,
                    #                             # center_adv=True,
                    #                             # positive_adv=False,
                    #                             # fixed_horizon=False,
                    #                             # pg_loss='surrogate_clip',
                    #                             lr_clip_range=1.0,
                    #                             max_kl_step=1.0,
                    #                             optimizer=optimizer,
                    #                             optimizer_args=optimizer_args,
                    #                             # policy_ent_coeff=0.0,
                    #                             # use_softplus_entropy=False,
                    #                             # use_neg_logli_entropy=False,
                    #                             # stop_entropy_gradient=False,
                    #                             # entropy_method='no_entropy',
                    #                             # name='PPO',
                    #                             )
                    #
                    #
                    #         local_runner.setup(algo=robust_algo,
                    #                      env=env,
                    #                      sampler_cls=sampler_cls,
                    #                      sampler_args=sampler_args)
                    #
                    #         results = local_runner.train(n_epochs=n_itr_robust, batch_size=batch_size_robust, plot=False)
                    #         # pdb.set_trace()
                    #         print('done')
                    #         with open(log_dir + '/paths.gz', 'wb') as f:
                    #             try:
                    #                 compress_pickle.dump(results, f, compression="gzip", set_default_extension=False)
                    #             except MemoryError:
                    #                 pdb.set_trace()

                    # saver = tf.train.Saver()
                    # save_path = saver.save(sess, log_dir + '/model.ckpt')
                    # print("Model saved in path: %s" % save_path)
                    #
                    # # Write out the episode results
                    # header = 'trial, step, ' + 'v_x_car, v_y_car, x_car, y_car, '
                    # for i in range(0,sim.c_num_peds):
                    #     header += 'v_x_ped_' + str(i) + ','
                    #     header += 'v_y_ped_' + str(i) + ','
                    #     header += 'x_ped_' + str(i) + ','
                    #     header += 'y_ped_' + str(i) + ','
                    #
                    # for i in range(0,sim.c_num_peds):
                    #     header += 'a_x_'  + str(i) + ','
                    #     header += 'a_y_' + str(i) + ','
                    #     header += 'noise_v_x_' + str(i) + ','
                    #     header += 'noise_v_y_' + str(i) + ','
                    #     header += 'noise_x_' + str(i) + ','
                    #     header += 'noise_y_' + str(i) + ','
                    #
                    # header += 'reward'
                    # if snapshot_mode != "gap":
                    #     snapshot_gap = n_itr - 1
                    # example_save_trials(n_itr, log_dir, header, sess, save_every_n=snapshot_gap)

    run_experiment(
        run_task,
        **run_experiment_args,
        # snapshot_mode=snapshot_mode,
        # log_dir=log_dir,
        # exp_name=exp_name,
        # snapshot_gap=snapshot_gap,
        # seed=1,
        # n_parallel=n_parallel,
    )


if __name__ == '__main__':
    fire.Fire()



class Baseline:
    def __init__(self, env_spec):
        self._mdp_spec = env_spec

    @property
    def algorithm_parallelized(self):
        return False

    def get_param_values(self):
        raise NotImplementedError

    def set_param_values(self, val):
        raise NotImplementedError

    def fit(self, paths):
        raise NotImplementedError

    def predict(self, path):
        raise NotImplementedError

    def log_diagnostics(self, paths):
        """
        Log extra information per iteration based on the collected paths
        """
        pass



import gym.utils.seeding as seeding
import numpy as np
from dowel import logger
from dowel import tabular
from garage.misc import tensor_utils as np_tensor_utils
from garage.tf.algos.batch_polopt import BatchPolopt
from garage.tf.misc import tensor_utils


class GA(BatchPolopt):
    """Deep Genetic Algorithm from Such et al.  [1]_.

    Parameters
    ----------
    top_paths : :py:class:`ast_toolbox.mcts.BoundedPriorityQueues`, optional
        The bounded priority queue to store top-rewarded trajectories.
    step_size : float, optional
        Standard deviation for each mutation.
    step_size_anneal : float, optional
        The linear annealing rate of step_size after each iteration.
    pop_size : int, optional
        The population size
    truncation_size : int, optional
        The number of top-performed individuals that are chosen as parents.
    keep_best : int, optional
        The number of top-performed individuals that remain unchanged for next generation.
    f_F : string, optional
        The method used to calculate fitness: 'mean' for the average return, 'max' for the max return.
    log_interval : int, optional
        The log interval in terms of environment calls.
    kwargs :
        Keyword arguments passed to :doc:`garage.tf.algos.BatchPolopt <garage:_apidoc/garage.tf.algos>`.

    References
    ----------
    .. [1] Such, Felipe Petroski, et al. "Deep neuroevolution: Genetic algorithms are a competitive
    alternative for training deep neural networks for reinforcement learning."
     arXiv:1712.06567 (2017).
    """

    def __init__(
            self,
            top_paths=None,
            n_itr=2,
            batch_size=500,
            step_size=0.01,
            step_size_anneal=1.0,
            pop_size=5,
            truncation_size=2,
            keep_best=1,
            f_F="mean",
            log_interval=4000,
            init_step=1.0,
            **kwargs):

        self.top_paths = top_paths
        self.best_mean = -np.inf
        self.best_var = 0.0
        self.n_itr = n_itr
        self.batch_size = batch_size
        self.step_size = step_size
        self.step_size_anneal = step_size_anneal
        self.pop_size = pop_size
        self.truncation_size = truncation_size
        self.keep_best = keep_best
        self.f_F = f_F
        self.log_interval = log_interval
        self.init_step = init_step

        self.seeds = np.zeros([n_itr, pop_size], dtype=int)
        self.magnitudes = np.zeros([n_itr, pop_size])
        self.parents = np.zeros(pop_size, dtype=int)
        self.np_random, seed = seeding.np_random()  # used in set_params
        super(GA, self).__init__(**kwargs)

    def initial(self):
        """Initiate trainer internal parameters.
        """
        self.seeds[0, :] = np.random.randint(low=0, high=int(2**16),
                                             size=(1, self.pop_size))
        self.magnitudes[0, :] = self.init_step * np.ones(self.pop_size)
        self.policy.set_param_values(self.policy.get_param_values())
        self.stepNum = 0

    def init_opt(self):
        """Initiate trainer internal tensorflow operations.
        """
        return dict()

    def train(self, runner):
        """Start training.

        Parameters
        ----------
        runner : :py:class:`garage.experiment.LocalRunner <garage:garage.experiment.LocalRunner>`
            ``LocalRunner`` is passed to give algorithm the access to ``runner.step_epochs()``, which provides services
            such as snapshotting and sampler control.
        """
        self.initial()

        for itr in runner.step_epochs():
            all_paths = {}
            for p in range(self.pop_size):
                with logger.prefix('idv #%d | ' % p):
                    logger.log("Updating Params")
                    self.set_params(itr, p)
                    logger.log("Obtaining samples...")
                    paths = self.obtain_samples(itr, runner)
                    logger.log("Processing samples...")
                    samples_data = self.process_samples(itr, paths)

                    # all_paths[p]=paths
                    all_paths[p] = samples_data

                    # logger.log("Logging diagnostics...")
                    # self.log_diagnostics(paths)

            logger.log("Optimizing Population...")
            self.optimize_policy(itr, all_paths)
            self.step_size = self.step_size * self.step_size_anneal
            self.record_tabular(itr)
            runner.step_itr += 1
        return None

    def record_tabular(self, itr):
        """Record training performace per-iteration.

        Parameters
        ----------
        itr : int
            The iteration number.
        """
        tabular.record('Itr', itr)
        tabular.record('StepNum', self.stepNum)
        # This causes tabular logging inconsistant
        # if self.top_paths is not None:
        #     for (topi, path) in enumerate(self.top_paths):
        #         tabular.record('reward ' + str(topi), path[0])
        tabular.record('BestMean', self.best_mean)
        tabular.record('BestVar', self.best_var)
        tabular.record('StepSize', self.step_size)
        tabular.record('Max Magnitude', np.max(self.magnitudes[itr, :]))
        tabular.record('Min Magnitude', np.min(self.magnitudes[itr, :]))
        tabular.record('Mean Magnitude', np.mean(self.magnitudes[itr, :]))
        self.extra_recording(itr)

    def extra_recording(self, itr):
        """Record extra training statistics per-iteration.

        Parameters
        ----------
        itr : int
            The iteration number.
        """
        return None

    def set_params(self, itr, p):
        """Set the current policy paramter to the specified iteration and individual.

        Parameters
        ----------
        itr : int
            The iteration number.
        p : int
            The individual index.
        """
        for i in range(itr + 1):
            # print("seed: ", self.seeds[i,p])
            self.np_random.seed(int(self.seeds[i, p]))
            if i == 0:  # first generation
                param_values = self.policy.get_param_values(trainable=True)
                param_values = self.magnitudes[i, p] * self.np_random.normal(size=param_values.shape)

                # param_values = init_policy_np(self.policy, self.np_random)

                # params = self.policy.get_params()
                # sess = tf.get_default_session()
                # sess.run(tf.variables_initializer(params))
                # param_values = self.policy.get_param_values()
            elif self.seeds[i, p] != 0:
                param_values = param_values + self.magnitudes[i, p] * self.np_random.normal(size=param_values.shape)
        self.policy.set_param_values(param_values, trainable=True)

    def get_fitness(self, itr, all_paths):
        """Calculate the fitness of the collexted paths.

        Parameters
        ----------
        itr : int
            The iteration number.
        all_paths : list[dict]
            The collected paths from the sampler.

        Returns
        ----------
        fitness : list[float]
            The list of fitness of each individual.
        """
        fitness = np.zeros(self.pop_size)
        for p in range(self.pop_size):
            rewards = all_paths[p]["rewards"]
            valid_rewards = rewards * all_paths[p]["valids"]
            path_rewards = np.sum(valid_rewards, -1)
            if self.f_F == "max":
                fitness[p] = np.max(path_rewards)
            else:
                fitness[p] = np.mean(path_rewards)
        return fitness

    def select_parents(self, fitness):
        """Select the individuals to be the parents of the next generation.

        Parameters
        ----------
        fitness : list[float]
            The list of fitness of each individual.
        """
        sort_indx = np.flip(np.argsort(fitness), axis=0)
        self.parents[0:self.truncation_size] = sort_indx[0:self.truncation_size]
        self.parents[self.truncation_size:self.pop_size] = \
            sort_indx[np.random.randint(low=0, high=self.truncation_size, size=self.pop_size - self.truncation_size)]

    def mutation(self, itr, new_seeds, new_magnitudes, all_paths):
        """Generate new random seeds and magnitudes for the next generation.

        The first self.keep_best seeds are set to no-mutation value (0).

        Parameters
        ----------
        itr : int
            The iteration number.
        new_seeds : :py:class:`numpy.ndarry`
            The original seeds.
        new_magnitudes : :py:class:`numpy.ndarry`
            The original magnitudes.
        all_paths : list[dict]
            The collected paths from the sampler.

        Returns
        -------
        new_seeds : :py:class:`numpy.ndarry`
            The new seeds.
        new_magnitudes : :py:class:`numpy.ndarry`
            The new magnitudes.
        """
        if itr + 1 < self.n_itr:
            new_seeds[itr + 1, :] = np.random.randint(low=0, high=int(2**32),
                                                      size=(1, self.pop_size))
            new_magnitudes[itr + 1, :] = self.step_size
            for i in range(0, self.keep_best):
                new_seeds[itr + 1, i] = 0
        return new_seeds, new_magnitudes

    def optimize_policy(self, itr, all_paths):
        """Update the population represented by self.seeds and self.parents.

        Parameters
        ----------
        itr : int
            The iteration number.
        all_paths : list[dict]
            The collected paths from the sampler.
        """
        fitness = self.get_fitness(itr, all_paths)
        self.select_parents(fitness)
        new_seeds = np.zeros_like(self.seeds)
        new_seeds[:, :] = self.seeds[:, self.parents]
        new_magnitudes = np.zeros_like(self.magnitudes)
        new_magnitudes[:, :] = self.magnitudes[:, self.parents]
        if itr + 1 < self.n_itr:
            new_seeds, new_magnitudes = self.mutation(itr, new_seeds, new_magnitudes, all_paths)
        self.seeds = new_seeds
        self.magnitudes = new_magnitudes
        return dict()

    def obtain_samples(self, itr, runner):
        """Collect rollout samples using the current policy paramter.

        Parameters
        ----------
        itr : int
            The iteration number.
        runner : :py:class:`garage.experiment.LocalRunner <garage:garage.experiment.LocalRunner>`
            ``LocalRunner`` is passed to give algorithm the access to ``runner.obtain_samples()``,
            which collects rollout paths from the sampler.

        Returns
        -------
        paths : list[dict]
            The collected paths from the sampler.
        """
        self.stepNum += self.batch_size
        # paths = self.sampler.obtain_samples(itr)
        paths = runner.obtain_samples(runner.step_itr)
        undiscounted_returns = [sum(path["rewards"]) for path in paths]
        if np.mean(undiscounted_returns) > self.best_mean:
            self.best_mean = np.mean(undiscounted_returns)
            self.best_var = np.var(undiscounted_returns)
        if not (self.top_paths is None):
            action_seqs = [path["actions"] for path in paths]
            [self.top_paths.enqueue(
                action_seq, R, make_copy=True) for (action_seq, R) in zip(action_seqs, undiscounted_returns)]
        return paths

    def process_samples(self, itr, paths):
        """Return processed sample data based on the collected paths.

        Parameters
        ----------
        itr : int
            The iteration number.
        paths : list[dict]
            The collected paths from the sampler.

        Returns
        -------
        samples_data : dict
            Processed sample data with same trajectory length (padded with 0)
        """
        baselines = []
        returns = []

        max_path_length = self.max_path_length

        if self.flatten_input:
            paths = [
                dict(
                    observations=(self.env_spec.observation_space.flatten_n(
                        path['observations'])),
                    actions=(
                        self.env_spec.action_space.flatten_n(  # noqa: E126
                            path['actions'])),
                    rewards=path['rewards'],
                    env_infos=path['env_infos'],
                    agent_infos=path['agent_infos']) for path in paths
            ]
        else:
            paths = [
                dict(
                    observations=path['observations'],
                    actions=(
                        self.env_spec.action_space.flatten_n(  # noqa: E126
                            path['actions'])),
                    rewards=path['rewards'],
                    env_infos=path['env_infos'],
                    agent_infos=path['agent_infos']) for path in paths
            ]

        if hasattr(self.baseline, 'predict_n'):
            all_path_baselines = self.baseline.predict_n(paths)
        else:
            all_path_baselines = [
                self.baseline.predict(path) for path in paths
            ]

        for idx, path in enumerate(paths):
            path_baselines = np.append(all_path_baselines[idx], 0)
            deltas = (path['rewards'] + self.discount * path_baselines[1:] -
                      path_baselines[:-1])
            path['advantages'] = np_tensor_utils.discount_cumsum(
                deltas, self.discount * self.gae_lambda)
            path['deltas'] = deltas

        for idx, path in enumerate(paths):
            # baselines
            path['baselines'] = all_path_baselines[idx]
            baselines.append(path['baselines'])

            # returns
            path['returns'] = np_tensor_utils.discount_cumsum(
                path['rewards'], self.discount)
            returns.append(path['returns'])

        # make all paths the same length
        obs = [path['observations'] for path in paths]
        obs = tensor_utils.pad_tensor_n(obs, max_path_length)

        actions = [path['actions'] for path in paths]
        actions = tensor_utils.pad_tensor_n(actions, max_path_length)

        rewards = [path['rewards'] for path in paths]
        rewards = tensor_utils.pad_tensor_n(rewards, max_path_length)

        returns = [path['returns'] for path in paths]
        returns = tensor_utils.pad_tensor_n(returns, max_path_length)

        baselines = tensor_utils.pad_tensor_n(baselines, max_path_length)

        agent_infos = [path['agent_infos'] for path in paths]
        agent_infos = tensor_utils.stack_tensor_dict_list([
            tensor_utils.pad_tensor_dict(p, max_path_length)
            for p in agent_infos
        ])

        env_infos = [path['env_infos'] for path in paths]
        env_infos = tensor_utils.stack_tensor_dict_list([
            tensor_utils.pad_tensor_dict(p, max_path_length) for p in env_infos
        ])

        valids = [np.ones_like(path['returns']) for path in paths]
        valids = tensor_utils.pad_tensor_n(valids, max_path_length)

        # average_discounted_return = (np.mean(
        #     [path['returns'][0] for path in paths]))

        undiscounted_returns = [sum(path['rewards']) for path in paths]
        self.episode_reward_mean.extend(undiscounted_returns)

        # ent = np.sum(self.policy.distribution.entropy(agent_infos) *
        #              valids) / np.sum(valids)

        samples_data = dict(
            observations=obs,
            actions=actions,
            rewards=rewards,
            baselines=baselines,
            returns=returns,
            valids=valids,
            agent_infos=agent_infos,
            env_infos=env_infos,
            paths=paths,
            average_return=np.mean(undiscounted_returns),
        )

        return samples_data

    def get_itr_snapshot(self, itr, samples_data):
        """Get the snapshot of the current population.

        Parameters
        ----------
        itr : int
            The iteration number.
        samples_data : dict
            The processed data samples.

        Returns
        -------
        snaposhot : dict
            The training snapshot.
        """
        return dict(
            itr=itr,
            policy=self.policy,
            seeds=self.seeds,
            magnitudes=self.magnitudes,
        )



import numpy as np

from ast_toolbox.mcts.AdaptiveStressTesting import AdaptiveStressTest
from ast_toolbox.mcts.AdaptiveStressTesting import ASTAction


class AdaptiveStressTestBV(AdaptiveStressTest):
    """The AST wrapper for MCTS using the Blind Value exploration [1]_.

    Parameters
    ----------
    kwargs :
        Keyword arguments passed to `ast_toolbox.mcts.AdaptiveStressTesting.AdaptiveStressTest`

    References
    ----------
    .. [1] Couetoux, Adrien, Hassen Doghmen, and Olivier Teytaud. "Improving the exploration in upper confidence trees."
     International Conference on Learning and Intelligent Optimization. Springer, Berlin, Heidelberg, 2012.
    """

    def __init__(self, **kwargs):
        super(AdaptiveStressTestBV, self).__init__(**kwargs)

    def explore_action(self, s, tree):
        """Sample an action for the exploration using Blind Value.

        Parameters
        ----------
        s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
            The current state.
        tree : dict
            The searching tree.

        Returns
        ----------
        action : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`
            The sampled action.
        """
        s = tree[s]
        A_explored = s.a.keys()
        if len(A_explored) == 0.0:
            return ASTAction(self.env.action_space.sample())
        UCB = self.getUCB(s)
        sigma_known = np.std([float(UCB[a]) for a in s.a.keys()])

        A_pool = []
        dist_pool = []
        center = (self.env.action_space.low + self.env.action_space.high) / 2.0
        for i in range(self.params.M):
            a = self.env.action_space.sample()
            A_pool.append(a)
            dist = self.getDistance(a, center)
            dist_pool.append(dist)
        sigma_pool = np.std(dist_pool)

        rho = sigma_known / sigma_pool

        BV_max = -np.inf
        a_best = None
        for y in A_pool:
            BV = self.getBV(y, rho, A_explored, UCB)
            if BV > BV_max:
                BV_max = BV
                a_best = y
        return ASTAction(a_best)

    def getDistance(self, a, b):
        """Get the (L2) distance between two actions.

        Parameters
        ----------
        a : :py:class:`numpy.ndarry`
            The first action.
        b : :py:class:`numpy.ndarry`
            The second action.

        Returns
        ----------
        distance : float
            The L2 distance between a and b.
        """
        return np.sqrt(np.sum((a - b)**2))

    def getUCB(self, s):
        """Get the upper confidnece bound for the expected return for evary actions that has been explored at the state.

        Parameters
        ----------
        s : :py:class:`ast_toolbox.MCTSdpw.StateNode`
            The state node in the searching tree

        Returns
        ----------
        UCB : dict
            The dictionary containing the upper confidence bound for each explored action in the state node.
        """
        UCB = dict()
        nS = s.n
        for a in s.a.keys():
            UCB[a] = s.a[a].q + self.params.ec * np.sqrt(np.log(nS) / float(s.a[a].n))
        return UCB

    def getBV(self, y, rho, A, UCB):
        """Calculate the Blind Value for the candidate action y

        Parameters
        ----------
        y : :py:class:`numpy.ndarry`
            The candidate action.
        rho : float
            The standard deviation ratio.
        A : list[:py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`]
            The list of the explored AST actions
        UCB : dict
            The dictionary containing the upper confidence bound for each explored action in the state node.

        Returns
        ----------
        BV : float
            The blind value.
        """
        BVs = []
        for a in A:
            BV = rho * self.getDistance(a.action, y) + UCB[a]
            BVs.append(BV)
        return min(BVs)



import pickle

from examples.AV.example_runner_ba_av import runner as ba_runner
from examples.AV.example_runner_drl_av import runner as drl_runner
from examples.AV.example_runner_ge_av import runner as go_explore_runner
from examples.AV.example_runner_mcts_av import runner as mcts_runner

if __name__ == '__main__':
    # Which algorithms to run
    RUN_DRL = False
    RUN_MCTS = True
    RUN_GA = False
    RUN_GE = False
    RUN_BA = False

    # Overall settings
    max_path_length = 50
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    base_log_dir = './data'
    # experiment settings
    run_experiment_args = {'snapshot_mode': 'last',
                           'snapshot_gap': 1,
                           'log_dir': None,
                           'exp_name': None,
                           'seed': 0,
                           'n_parallel': 8,
                           'tabular_log_file': 'progress.csv'
                           }

    # runner settings
    runner_args = {'n_epochs': 5,
                   'batch_size': 500,
                   'plot': False
                   }

    # env settings
    env_args = {'id': 'ast_toolbox:GoExploreAST-v1',
                'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_init_state': True,
                's_0': s_0,
                }

    # simulation settings
    sim_args = {'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_initial_state': True,
                'max_path_length': max_path_length
                }

    # reward settings
    reward_args = {'use_heuristic': True}

    # spaces settings
    spaces_args = {}

    # DRL ----------------------------------------------------------------------------------

    if RUN_DRL:
        # DRL Settings

        drl_policy_args = {'name': 'lstm_policy',
                           'hidden_dim': 64,
                           }

        drl_baseline_args = {}

        drl_algo_args = {'max_path_length': max_path_length,
                         'discount': 0.99,
                         'lr_clip_range': 1.0,
                         'max_kl_step': 1.0,
                         # 'log_dir':None,
                         }

        run_experiment_args['log_dir'] = base_log_dir + '/drl'
        run_experiment_args['exp_name'] = 'drl'

        drl_algo_args['max_path_length'] = max_path_length

        # Run DRL
        drl_runner(
            env_args=env_args,
            run_experiment_args=run_experiment_args,
            sim_args=sim_args,
            reward_args=reward_args,
            spaces_args=spaces_args,
            policy_args=drl_policy_args,
            baseline_args=drl_baseline_args,
            algo_args=drl_algo_args,
            runner_args=runner_args,
        )

    # MCTS ----------------------------------------------------------------------------------

    if RUN_MCTS:
        # MCTS Settings

        mcts_type = 'mcts'

        mcts_sampler_args = {}

        mcts_algo_args = {'max_path_length': max_path_length,
                          'stress_test_mode': 2,
                          'ec': 100.0,
                          'n_itr': 1,
                          'k': 0.5,
                          'alpha': 0.5,
                          'clear_nodes': True,
                          'log_interval': 500,
                          'plot_tree': False,
                          'plot_path': None,
                          'log_dir': None,
                          }

        mcts_bpq_args = {'N': 10}

        # MCTS settings
        run_experiment_args['log_dir'] = base_log_dir + '/mcts'
        run_experiment_args['exp_name'] = 'mcts'

        mcts_algo_args['max_path_length'] = max_path_length
        mcts_algo_args['log_dir'] = run_experiment_args['log_dir']
        mcts_algo_args['plot_path'] = run_experiment_args['log_dir']

        mcts_runner(
            mcts_type=mcts_type,
            env_args=env_args,
            run_experiment_args=run_experiment_args,
            sim_args=sim_args,
            reward_args=reward_args,
            spaces_args=spaces_args,
            algo_args=mcts_algo_args,
            runner_args=runner_args,
            bpq_args=mcts_bpq_args,
            sampler_args=mcts_sampler_args,
            save_expert_trajectory=True,
        )

    # Go-Explore ----------------------------------------------------------------------------------

    if RUN_GE:
        # Go-Explore Settings

        ge_algo_args = {'db_filename': None,
                        'max_db_size': 150,
                        'max_path_length': max_path_length,
                        'discount': 0.99,
                        'save_paths_gap': 1,
                        'save_paths_path': None,
                        'overwrite_db': True,
                        }

        ge_baseline_args = {}

        ge_policy_args = {}

        run_experiment_args['log_dir'] = base_log_dir + '/ge'
        run_experiment_args['exp_name'] = 'ge'
        # run_experiment_args['tabular_log_file'] = 'progress.csv'

        ge_algo_args['db_filename'] = run_experiment_args['log_dir'] + '/cellpool'
        ge_algo_args['save_paths_path'] = run_experiment_args['log_dir']
        ge_algo_args['max_path_length'] = max_path_length

        go_explore_runner(
            env_args=env_args,
            run_experiment_args=run_experiment_args,
            sim_args=sim_args,
            reward_args=reward_args,
            spaces_args=spaces_args,
            policy_args=ge_policy_args,
            baseline_args=ge_baseline_args,
            algo_args=ge_algo_args,
            runner_args=runner_args,
        )

    # Backward Algorithm ----------------------------------------------------------------------------------

    if RUN_BA:
        # BA Settings
        ba_algo_args = {'expert_trajectory': None,
                        'max_path_length': max_path_length,
                        'epochs_per_step': 10,
                        'scope': None,
                        'discount': 0.99,
                        'gae_lambda': 1.0,
                        'center_adv': True,
                        'positive_adv': False,
                        'fixed_horizon': False,
                        'pg_loss': 'surrogate_clip',
                        'lr_clip_range': 1.0,
                        'max_kl_step': 1.0,
                        'policy_ent_coeff': 0.0,
                        'use_softplus_entropy': False,
                        'use_neg_logli_entropy': False,
                        'stop_entropy_gradient': False,
                        'entropy_method': 'no_entropy',
                        'name': 'PPO',
                        'log_dir': None,
                        }

        ba_baseline_args = {}

        ba_policy_args = {'name': 'lstm_policy',
                          'hidden_dim': 64,
                          }

        with open(run_experiment_args['log_dir'] + '/expert_trajectory.p', 'rb') as f:
            expert_trajectories = pickle.load(f)

        run_experiment_args['log_dir'] = run_experiment_args['log_dir'] + '/ba'
        ba_algo_args['expert_trajectory'] = expert_trajectories[-1]

        ba_runner(
            env_args=env_args,
            run_experiment_args=run_experiment_args,
            sim_args=sim_args,
            reward_args=reward_args,
            spaces_args=spaces_args,
            policy_args=ba_policy_args,
            baseline_args=ba_baseline_args,
            algo_args=ba_algo_args,
            runner_args=runner_args,
        )



#!/usr/bin/env python3
"""Example using VPG with ISSampler.

Iterations alternate between live and importance sampled iterations.
"""
import gym

from garage.envs import normalize
from garage.experiment import run_experiment
from garage.np.baselines import LinearFeatureBaseline
from garage.sampler import ISSampler
from garage.tf.algos import VPG
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import GaussianMLPPolicy


def run_task(snapshot_config, *_):
    """Run the job.

    Args:
        snapshot_config (garage.experiment.SnapshotConfig): Configuration
            values for snapshotting.
        *_ (object): Hyperparameters (unused).

    """
    with LocalTFRunner(snapshot_config=snapshot_config) as runner:
        env = TfEnv(normalize(gym.make('InvertedPendulum-v2')))

        policy = GaussianMLPPolicy(env_spec=env.spec, hidden_sizes=(32, 32))

        baseline = LinearFeatureBaseline(env_spec=env.spec)

        algo = VPG(
            env_spec=env.spec,
            policy=policy,
            baseline=baseline,
            max_path_length=100,
            discount=0.99,
            max_kl_step=0.01,
        )

        runner.setup(algo,
                     env,
                     sampler_cls=ISSampler,
                     sampler_args=dict(n_backtrack=1))
        runner.train(n_epochs=40, batch_size=4000)


run_experiment(
    run_task,
    snapshot_mode='last',
    seed=1,
)



from tests.fixtures.envs.dummy import DummyBoxEnv


class DummyRewardBoxEnv(DummyBoxEnv):
    """A dummy box environment."""

    def __init__(self, random=True):
        super().__init__(random)

    def step(self, action):
        """Step the environment."""
        if action == 0:
            reward = 10
        else:
            reward = -10
        return self.observation_space.sample(), reward, True, dict()



#!/usr/bin/env python3
"""This is an example to train multiple tasks with PPO algorithm."""
import gym
import tensorflow as tf

from garage.envs import normalize
from garage.envs.multi_env_wrapper import MultiEnvWrapper
from garage.experiment import run_experiment
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import PPO
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import CategoricalMLPPolicy


def run_task(snapshot_config, *_):
    """Run task.

    Args:
        snapshot_config (garage.experiment.SnapshotConfig): The snapshot
            configuration used by LocalRunner to create the snapshotter.

        _ (object): Ignored by this function.

    """
    with LocalTFRunner(snapshot_config=snapshot_config) as runner:
        env1 = TfEnv(normalize(gym.make('Adventure-ram-v4')))
        env2 = TfEnv(normalize(gym.make('Alien-ram-v4')))
        env = MultiEnvWrapper([env1, env2])

        policy = CategoricalMLPPolicy(
            env_spec=env.spec,
            hidden_nonlinearity=tf.nn.tanh,
        )

        baseline = LinearFeatureBaseline(env_spec=env.spec)

        algo = PPO(env_spec=env.spec,
                   policy=policy,
                   baseline=baseline,
                   max_path_length=100,
                   discount=0.99,
                   gae_lambda=0.95,
                   lr_clip_range=0.2,
                   policy_ent_coeff=0.0,
                   optimizer_args=dict(
                       batch_size=32,
                       max_epochs=10,
                       tf_optimizer_args=dict(learning_rate=1e-3),
                   ))

        runner.setup(algo, env)
        runner.train(n_epochs=120, batch_size=2048, plot=False)


run_experiment(run_task, snapshot_mode='last', seed=1)



import ast_toolbox.mcts.MCTSdpw as MCTSdpw
import ast_toolbox.mcts.MDP as MDP


def rollout_getAction(ast):
    """Get the rollout function from ast.

    Parameters
    ----------
    ast : :py:class:`ast_toolbox.mcts.AdaptiveStressTest.AdaptiveStressTesting`
        The AST object.
    """
    def rollout_policy(s, tree):
        return ast.random_action()
    return rollout_policy


def explore_getAction(ast):
    """Get the exploration function from ast.

    Parameters
    ----------
    ast : :py:class:`ast_toolbox.mcts.AdaptiveStressTest.AdaptiveStressTesting`
        The AST object.
    """
    def explore_policy(s, tree):
        return ast.explore_action(s, tree)
    return explore_policy


def stress_test(ast, mcts_params, top_paths, verbose=True, return_tree=False):
    """Run stress test with mode 1 (search with single tree).

    Parameters
    ----------
    ast : :py:class:`ast_toolbox.mcts.AdaptiveStressTest.AdaptiveStressTesting`
        The AST object.
    mcts_params: :py:class:`ast_toolbox.mcts.MCTSdpw.DPWParams`
        The mcts parameters.
    top_paths : :py:class:`ast_toolbox.mcts.BoundedPriorityQueues`
        The bounded priority queue to store top-rewarded trajectories.
    verbose : bool, optional
        Whether to logging test information
    return_tree: bool, optional
        Whether to return the search tree

    Returns
    -------
    results : :py:class:`ast_toolbox.mcts.AdaptiveStressTest.AdaptiveStressTesting`
        The bounded priority queue storing top-rewarded trajectories.
    tree : dict
        The resulting searching tree.
    """
    dpw_model = MCTSdpw.DPWModel(ast.transition_model, rollout_getAction(ast), explore_getAction(ast))
    tree = MCTSdpw.DPWTree(mcts_params, dpw_model)
    (mcts_reward, action_seq) = MDP.simulate(tree.f.model, tree, MCTSdpw.selectAction, verbose=verbose)

    results = ast.top_paths
    if return_tree:
        return results, tree.s_tree
    else:
        return results


def stress_test2(ast, mcts_params, top_paths, verbose=True, return_tree=False):
    """Run stress test with mode 2 (search with multiple trees).

    Parameters
    ----------
    ast : :py:class:`ast_toolbox.mcts.AdaptiveStressTest.AdaptiveStressTesting`
        The AST object.
    mcts_params: :py:class:`ast_toolbox.mcts.MCTSdpw.DPWParams`
        The mcts parameters.
    top_paths : :py:class:`ast_toolbox.mcts.BoundedPriorityQueues`
        The bounded priority queue to store top-rewarded trajectories.
    verbose : bool, optional
        Whether to logging test information
    return_tree: bool, optional
        Whether to return the search tree

    Returns
    -------
    results : :py:class:`ast_toolbox.mcts.AdaptiveStressTest.AdaptiveStressTesting`
        The bounded priority queue storing top-rewarded trajectories.
    tree : dict
        The resulting searching tree.
    """
    mcts_params.clear_nodes = False
    mcts_params.n *= ast.params.max_steps

    dpw_model = MCTSdpw.DPWModel(ast.transition_model, rollout_getAction(ast), explore_getAction(ast))
    tree = MCTSdpw.DPWTree(mcts_params, dpw_model)

    s = tree.f.model.getInitialState()
    MCTSdpw.selectAction(tree, s, verbose=verbose)

    results = ast.top_paths
    if return_tree:
        return results, tree.s_tree
    else:
        return results



"""Cross Entropy Method."""
from dowel import logger, tabular
import numpy as np

from garage.np.algos import BatchPolopt


class CEM(BatchPolopt):
    """Cross Entropy Method.

    CEM works by iteratively optimizing a gaussian distribution of policy.

    In each epoch, CEM does the following:
    1. Sample n_samples policies from a gaussian distribution of
       mean cur_mean and std cur_std.
    2. Do rollouts for each policy.
    3. Update cur_mean and cur_std by doing Maximum Likelihood Estimation
       over the n_best top policies in terms of return.

    Note:
        When training CEM with LocalRunner, make sure that n_epoch_cycles for
        runner equals to n_samples for CEM.

        This implementation leverage n_epoch_cycles to do rollouts for a single
        policy in an epoch cycle.

    Args:
        env_spec (garage.envs.EnvSpec): Environment specification.
        policy (garage.np.policies.Policy): Action policy.
        baseline(garage.np.baselines.Baseline): Baseline for GAE
            (Generalized Advantage Estimation).
        n_samples (int): Number of policies sampled in one epoch.
        discount (float): Environment reward discount.
        max_path_length (int): Maximum length of a single rollout.
        best_frac (float): The best fraction.
        init_std (float): Initial std for policy param distribution.
        extra_std (float): Decaying std added to param distribution.
        extra_decay_time (float): Epochs that it takes to decay extra std.

    """

    def __init__(self,
                 env_spec,
                 policy,
                 baseline,
                 n_samples,
                 discount=0.99,
                 max_path_length=500,
                 init_std=1,
                 best_frac=0.05,
                 extra_std=1.,
                 extra_decay_time=100):
        super().__init__(policy, baseline, discount, max_path_length,
                         n_samples)
        self.env_spec = env_spec

        self.n_samples = n_samples
        self.best_frac = best_frac
        self.init_std = init_std
        self.best_frac = best_frac
        self.extra_std = extra_std
        self.extra_decay_time = extra_decay_time

    def _sample_params(self, epoch):
        extra_var_mult = max(1.0 - epoch / self.extra_decay_time, 0)
        sample_std = np.sqrt(
            np.square(self.cur_std) +
            np.square(self.extra_std) * extra_var_mult)
        return np.random.standard_normal(
            self.n_params) * sample_std + self.cur_mean

    def train(self, runner):
        """Initialize variables and start training.

        Args:
            runner (LocalRunner): LocalRunner is passed to give algorithm
                the access to runner.step_epochs(), which provides services
                such as snapshotting and sampler control.

        Returns:
            The average return in last epoch cycle.

        """
        # epoch-wise
        self.cur_std = self.init_std
        self.cur_mean = self.policy.get_param_values()
        # epoch-cycle-wise
        self.cur_params = self.cur_mean
        self.all_returns = []
        self.all_params = [self.cur_mean.copy()]
        # constant
        self.n_best = int(self.n_samples * self.best_frac)
        assert self.n_best >= 1, (
            'n_samples is too low. Make sure that n_samples * best_frac >= 1')
        self.n_params = len(self.cur_mean)

        return super().train(runner)

    def train_once(self, itr, paths):
        """Perform one step of policy optimization given one batch of samples.

        Args:
            itr (int): Iteration number.
            paths (list[dict]): A list of collected paths.

        """
        paths = self.process_samples(itr, paths)

        epoch = itr // self.n_samples
        i_sample = itr - epoch * self.n_samples
        tabular.record('Epoch', epoch)
        tabular.record('# Sample', i_sample)
        # -- Stage: Process path
        rtn = paths['average_return']
        self.all_returns.append(paths['average_return'])

        # -- Stage: Update policy distribution.
        if (itr + 1) % self.n_samples == 0:
            avg_rtns = np.array(self.all_returns)
            best_inds = np.argsort(-avg_rtns)[:self.n_best]
            best_params = np.array(self.all_params)[best_inds]

            # MLE of normal distribution
            self.cur_mean = best_params.mean(axis=0)
            self.cur_std = best_params.std(axis=0)
            self.policy.set_param_values(self.cur_mean)

            # Clear for next epoch
            rtn = max(self.all_returns)
            self.all_returns.clear()
            self.all_params.clear()

        # -- Stage: Generate a new policy for next path sampling
        self.cur_params = self._sample_params(itr)
        self.all_params.append(self.cur_params.copy())
        self.policy.set_param_values(self.cur_params)

        logger.log(tabular)
        return rtn



import csv
import os.path

from matplotlib import pyplot as plt

n_trial = 5
top_k = 10
batch_size = 4000
# max_step = 5e6

prepath = "../"
exps = ["CartpoleNdRewardt"]
policies = ["GATRDInterStep1.0Fmax", "PSMCTSInterStep1.0Ec1.414K0.5A0.5Qmax",
            "PSMCTSTRInterStep1.0Ec1.414K0.5A0.5Qmax", "PSMCTSTRCInterStep1.0Ec1.414K0.5A0.5Qmax"]
plot_name = "GATRD_PSMCTS"
# colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']
colors = []
for i in range(len(policies)):
    colors.append('C' + str(i))

for exp in exps:
    plts = []
    legends = []
    fig = plt.figure(figsize=(10, 10))
    for (policy_index, policy) in enumerate(policies):
        print(policy)
        for trial in range(n_trial):
            print(trial)
            steps = []
            rewards = []
            file_path = prepath + exp + '/Data/AST/Lexington/' + policy + '/' + str(trial) + '/process.csv'
            if os.path.exists(file_path):
                with open(file_path) as csv_file:
                    csv_reader = csv.reader(csv_file, delimiter=',')
                    for (i, row) in enumerate(csv_reader):
                        if i == 0:
                            entry_dict = {}
                            for index in range(len(row)):
                                entry_dict[row[index]] = index
                        else:
                            # if int(row[entry_dict["StepNum"]]) > max_step:
                            # break
                            if int(row[entry_dict["StepNum"]]) % batch_size == 0:
                                steps.append(int(row[entry_dict["StepNum"]]))
                                rewards.append(max(0.0, float(row[entry_dict["reward 0"]])))
                    plot, = plt.plot(steps, rewards, color=colors[policy_index])
                    plts.append(plot)
                    legends.append(policy + ' ' + str(trial))

    plt.legend(plts, legends)
    plt.xlabel('Step number')
    plt.ylabel('Best reward')
    fig.savefig(prepath + exp + '/Data/Plot/top1/' + plot_name + '_top1.pdf')
    plt.close(fig)



#!/usr/bin/env python3
"""Run an experiment triggered by `run_experiment()` in `experiment.py`."""
import argparse
import ast
import base64
import datetime
import enum
import json
import os
import pathlib
import pickle
import sys
import uuid

import cloudpickle
import dateutil.tz
import dowel
from dowel import logger
import psutil

from garage.experiment import deterministic, SnapshotConfig
import garage.plotter
from garage.sampler import parallel_sampler
import garage.tf.plotter


def is_iterable(obj):
    """Return if instance is iterable."""
    return isinstance(obj, str) or getattr(obj, '__iter__', False)


def run_experiment(argv):
    """Run experiment."""
    now = datetime.datetime.now(dateutil.tz.tzlocal())

    # avoid name clashes when running distributed jobs
    rand_id = str(uuid.uuid4())[:5]
    timestamp = now.strftime('%Y_%m_%d_%H_%M_%S_%f_%Z')

    default_exp_name = 'experiment_%s_%s' % (timestamp, rand_id)
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--n_parallel',
        type=int,
        default=1,
        help=('Number of parallel workers to perform rollouts. '
              "0 => don't start any workers"))
    parser.add_argument('--exp_name',
                        type=str,
                        default=default_exp_name,
                        help='Name of the experiment.')
    parser.add_argument('--log_dir',
                        type=str,
                        default=None,
                        help='Path to save the log and iteration snapshot.')
    parser.add_argument('--snapshot_mode',
                        type=str,
                        default='last',
                        help='Mode to save the snapshot. Can be either "all" '
                        '(all iterations will be saved), "last" (only '
                        'the last iteration will be saved), "gap" (every'
                        '`snapshot_gap` iterations are saved), or "none" '
                        '(do not save snapshots)')
    parser.add_argument('--snapshot_gap',
                        type=int,
                        default=1,
                        help='Gap between snapshot iterations.')
    parser.add_argument(
        '--resume_from_dir',
        type=str,
        default=None,
        help='Directory of the pickle file to resume experiment from.')
    parser.add_argument('--resume_from_epoch',
                        type=str,
                        default=None,
                        help='Index of iteration to restore from. '
                        'Can be "first", "last" or a number. '
                        'Not applicable when snapshot_mode="last"')
    parser.add_argument('--tabular_log_file',
                        type=str,
                        default='progress.csv',
                        help='Name of the tabular log file (in csv).')
    parser.add_argument('--text_log_file',
                        type=str,
                        default='debug.log',
                        help='Name of the text log file (in pure text).')
    parser.add_argument('--tensorboard_step_key',
                        type=str,
                        default=None,
                        help='Name of the step key in tensorboard_summary.')
    parser.add_argument('--params_log_file',
                        type=str,
                        default='params.json',
                        help='Name of the parameter log file (in json).')
    parser.add_argument('--variant_log_file',
                        type=str,
                        default='variant.json',
                        help='Name of the variant log file (in json).')
    parser.add_argument('--plot',
                        type=ast.literal_eval,
                        default=False,
                        help='Whether to plot the iteration results')
    parser.add_argument(
        '--log_tabular_only',
        type=ast.literal_eval,
        default=False,
        help='Print only the tabular log information (in a horizontal format)')
    parser.add_argument('--seed',
                        type=int,
                        default=None,
                        help='Random seed for numpy')
    parser.add_argument('--args_data',
                        type=str,
                        help='Pickled data for objects')
    parser.add_argument('--variant_data',
                        type=str,
                        help='Pickled data for variant configuration')

    args = parser.parse_args(argv[1:])

    if args.seed is not None:
        deterministic.set_seed(args.seed)

    if args.n_parallel > 0:
        parallel_sampler.initialize(n_parallel=args.n_parallel)
        if args.seed is not None:
            parallel_sampler.set_seed(args.seed)

    if not args.plot:
        garage.plotter.Plotter.disable()
        garage.tf.plotter.Plotter.disable()

    if args.log_dir is None:
        log_dir = os.path.join(os.path.join(os.getcwd(), 'data'),
                               args.exp_name)
    else:
        log_dir = args.log_dir

    tabular_log_file = os.path.join(log_dir, args.tabular_log_file)
    text_log_file = os.path.join(log_dir, args.text_log_file)
    params_log_file = os.path.join(log_dir, args.params_log_file)

    if args.variant_data is not None:
        variant_data = pickle.loads(base64.b64decode(args.variant_data))
        variant_log_file = os.path.join(log_dir, args.variant_log_file)
        dump_variant(variant_log_file, variant_data)
    else:
        variant_data = None

    log_parameters(params_log_file, args)

    logger.add_output(dowel.TextOutput(text_log_file))
    logger.add_output(dowel.CsvOutput(tabular_log_file))
    logger.add_output(dowel.TensorBoardOutput(log_dir))
    logger.add_output(dowel.StdOutput())

    logger.push_prefix('[%s] ' % args.exp_name)

    snapshot_config = SnapshotConfig(snapshot_dir=log_dir,
                                     snapshot_mode=args.snapshot_mode,
                                     snapshot_gap=args.snapshot_gap)

    method_call = cloudpickle.loads(base64.b64decode(args.args_data))
    try:
        method_call(snapshot_config, variant_data, args.resume_from_dir,
                    args.resume_from_epoch)
    except BaseException:
        children = garage.plotter.Plotter.get_plotters()
        children += garage.tf.plotter.Plotter.get_plotters()
        if args.n_parallel > 0:
            children += [parallel_sampler]
        child_proc_shutdown(children)
        raise

    logger.remove_all()
    logger.pop_prefix()


def child_proc_shutdown(children):
    """Shut down children processes."""
    run_exp_proc = psutil.Process()
    alive = run_exp_proc.children(recursive=True)
    for proc in alive:
        if any([
                'multiprocessing.semaphore_tracker' in cmd
                for cmd in proc.cmdline()
        ]):
            alive.remove(proc)

    for c in children:
        c.close()
    max_retries = 5
    for _ in range(max_retries):
        _, alive = psutil.wait_procs(alive, 1.0)
        if not alive:
            break
    if alive:
        error_msg = ''
        for child in alive:
            error_msg += '{}\n'.format(
                str(
                    child.as_dict(
                        attrs=['ppid', 'pid', 'name', 'status', 'cmdline'])))

        error_msg = ("The following processes didn't die after the shutdown "
                     'of run_experiment:\n') + error_msg
        error_msg += ('This is a sign of an unclean shutdown. Please reopen '
                      'the following issue\nwith a detailed description '
                      'of how the error was produced:\n')
        error_msg += ('https://github.com/rlworkgroup/garage/issues/120')
        print(error_msg)


def log_parameters(log_file, args):
    """Log parameters to file."""
    log_params = {}
    for param_name, param_value in args.__dict__.items():
        log_params[param_name] = param_value
    if args.args_data is not None:
        log_params['json_args'] = dict()
    pathlib.Path(os.path.dirname(log_file)).mkdir(parents=True, exist_ok=True)
    with open(log_file, 'w') as f:
        json.dump(log_params, f, indent=2, sort_keys=True, cls=LogEncoder)


def dump_variant(log_file, variant_data):
    """Dump the variant file."""
    pathlib.Path(os.path.dirname(log_file)).mkdir(parents=True, exist_ok=True)
    with open(log_file, 'w') as f:
        json.dump(variant_data, f, indent=2, sort_keys=True, cls=LogEncoder)


class LogEncoder(json.JSONEncoder):
    """Encoder to be used as cls in json.dump."""

    def default(self, o):  # pylint: disable=method-hidden
        """Perform JSON encoding."""
        if isinstance(o, type):
            return {'$class': o.__module__ + '.' + o.__name__}
        elif isinstance(o, enum.Enum):
            return {
                '$enum':
                o.__module__ + '.' + o.__class__.__name__ + '.' + o.name
            }
        elif callable(o):
            return {'$function': o.__module__ + '.' + o.__name__}
        return json.JSONEncoder.default(self, o)


if __name__ == '__main__':
    run_experiment(sys.argv)



"""Episodic life wrapper for gym.Env."""
import gym


class EpisodicLife(gym.Wrapper):
    """Episodic life wrapper for gym.Env.

    This wrapper makes episode end when a life is lost, but only reset
    when all lives are lost.

    Args:
        env: The environment to be wrapped.
    """

    def __init__(self, env):
        super().__init__(env)
        self._lives = 0
        self._was_real_done = True

    def step(self, action):
        """gym.Env step function."""
        obs, reward, done, info = self.env.step(action)
        self._was_real_done = done
        lives = self.env.unwrapped.ale.lives()
        if lives < self._lives and lives > 0:
            done = True
        self._lives = lives
        return obs, reward, done, info

    def reset(self, **kwargs):
        """
        gym.Env reset function.

        Reset only when lives are lost.
        """
        if self._was_real_done:
            obs = self.env.reset(**kwargs)
        else:
            # no-op step
            obs, _, _, _ = self.env.step(0)
        self._lives = self.env.unwrapped.ale.lives()
        return obs



"""Class to define the action and observation spaces for an example AV validation task."""
import numpy as np
from gym.spaces.box import Box

from ast_toolbox.spaces import ASTSpaces


class ExampleAVSpaces(ASTSpaces):
    """Class to define the action and observation spaces for an example AV validation task.

    Parameters
    ----------
    num_peds : int, optional
        The number of pedestrians crossing the street.
    max_path_length : int, optional
        Maximum length of a single rollout.
    v_des : float, optional
        The desired velocity, in meters per second,  for the ego vehicle to maintain
    x_accel_low : float, optional
        The minimum x-acceleration of the pedestrian.
    y_accel_low : float, optional
        The minimum y-acceleration of the pedestrian.
    x_accel_high : float, optional
        The maximum x-acceleration of the pedestrian.
    y_accel_high : float, optional
        The maximum y-acceleration of the pedestrian.
    x_boundary_low : float, optional
        The minimum x-position of the pedestrian.
    y_boundary_low : float, optional
        The minimum y-position of the pedestrian.
    x_boundary_high : float, optional
        The maximum x-position of the pedestrian.
    y_boundary_high : float, optional
        The maximum y-position of the pedestrian.
    x_v_low : float, optional
        The minimum x-velocity of the pedestrian.
    y_v_low : float, optional
        The minimum y-velocity of the pedestrian.
    x_v_high : float, optional
        The maximum x-velocity of the pedestrian.
    y_v_high : float, optional
        The maximum y-velocity of the pedestrian.
    car_init_x : float, optional
        The initial x-position of the ego vehicle.
    car_init_y : float, optional
        The initial y-position of the ego vehicle.
    open_loop : bool, optional
        True if the simulation is open-loop, meaning that AST must generate all actions ahead of time, instead
        of being able to output an action in sync with the simulator, getting an observation back before
        the next action is generated. False to get interactive control, which requires that `blackbox_sim_state`
        is also False.
    """

    def __init__(self,
                 num_peds=1,
                 max_path_length=50,
                 v_des=11.17,
                 x_accel_low=-1.0,
                 y_accel_low=-1.0,
                 x_accel_high=1.0,
                 y_accel_high=1.0,
                 x_boundary_low=-10.0,
                 y_boundary_low=-10.0,
                 x_boundary_high=10.0,
                 y_boundary_high=10.0,
                 x_v_low=-10.0,
                 y_v_low=-10.0,
                 x_v_high=10.0,
                 y_v_high=10.0,
                 car_init_x=-35.0,
                 car_init_y=0.0,
                 open_loop=True,
                 ):
        # Constant hyper-params -- set by user
        self.c_num_peds = num_peds
        self.c_max_path_length = max_path_length
        self.c_v_des = v_des
        self.c_x_accel_low = x_accel_low
        self.c_y_accel_low = y_accel_low
        self.c_x_accel_high = x_accel_high
        self.c_y_accel_high = y_accel_high
        self.c_x_boundary_low = x_boundary_low
        self.c_y_boundary_low = y_boundary_low
        self.c_x_boundary_high = x_boundary_high
        self.c_y_boundary_high = y_boundary_high
        self.c_x_v_low = x_v_low
        self.c_y_v_low = y_v_low
        self.c_x_v_high = x_v_high
        self.c_y_v_high = y_v_high
        self.c_car_init_x = car_init_x
        self.c_car_init_y = car_init_y
        self.open_loop = open_loop
        self.low_start_bounds = [-1.0, -6.0, -1.0, 5.0, 0.0, -6.0, 0.0, 5.0]
        self.high_start_bounds = [1.0, -1.0, 0.0, 9.0, 1.0, -2.0, 1.0, 9.0]
        self.v_start = [1.0, -1.0, 1.0, -1.0]
        super().__init__()

    @property
    def action_space(self):
        """Returns a definition of the action space of the reinforcement learning problem.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The action space of the reinforcement learning problem.
        """
        low = np.array([self.c_x_accel_low, self.c_y_accel_low, -3.0, -3.0, -3.0, -3.0])
        high = np.array([self.c_x_accel_high, self.c_y_accel_high, 3.0, 3.0, 3.0, 3.0])

        for i in range(1, self.c_num_peds):
            low = np.hstack((low, np.array([self.c_x_accel_low, self.c_y_accel_low, 0.0, 0.0, 0.0, 0.0])))
            high = np.hstack((high, np.array([self.c_x_accel_high, self.c_y_accel_high, 1.0, 1.0, 1.0, 1.0])))

        return Box(low=low, high=high, dtype=np.float32)

    @property
    def observation_space(self):
        """Returns a definition of the observation space of the reinforcement learning problem.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The observation space of the reinforcement learning problem.
        """

        low = np.array([self.c_x_v_low, self.c_y_v_low, self.c_x_boundary_low, self.c_y_boundary_low])
        high = np.array([self.c_x_v_high, self.c_y_v_high, self.c_x_boundary_high, self.c_y_boundary_high])

        for i in range(1, self.c_num_peds):
            low = np.hstack(
                (low, np.array([self.c_x_v_low, self.c_y_v_low, self.c_x_boundary_low, self.c_y_boundary_low])))
            high = np.hstack(
                (high, np.array([self.c_x_v_high, self.c_y_v_high, self.c_x_boundary_high, self.c_y_boundary_high])))

        if self.open_loop:
            low = self.low_start_bounds[:self.c_num_peds * 2]
            low = low + np.ndarray.tolist(0.0 * np.array(self.v_start))[:self.c_num_peds]
            low = low + [0.75 * self.c_v_des]

            high = self.high_start_bounds[:self.c_num_peds * 2]
            high = high + np.ndarray.tolist(2.0 * np.array(self.v_start))[:self.c_num_peds]
            high = high + [1.25 * self.c_v_des]

            if self.c_car_init_x > 0:
                low = low + [0.75 * self.c_car_init_x]
                high = high + [1.25 * self.c_car_init_x]
            else:
                low = low + [1.25 * self.c_car_init_x]
                high = high + [0.75 * self.c_car_init_x]

        return Box(low=np.array(low), high=np.array(high), dtype=np.float32)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('./'))
# sys.path.insert(0, os.path.abspath('../'))
# sys.path.insert(0, os.path.abspath('../..'))
# sys.path.insert(0, os.path.abspath('../src/ast_toolbox/'))
# sys.path.insert(0, os.path.abspath('../src/'))
# sys.path.insert(0, os.path.abspath('../../src/ast_toolbox/'))
sys.path.insert(0, os.path.abspath('../../src/'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
# sphinx.ext.autodoc config
autodoc_mock_imports = ['bsddb3']
autodoc_member_order = 'groupwise'

source_suffix = '.rst'
master_doc = 'index'
project = 'AdaptiveStressTestingToolbox'
year = '2018-2020'
author = 'Stanford Intelligent Systems Laboratory'
copyright = '{0}, {1}'.format(year, author)
version = release = '2020.09.01.rc1'

pygments_style = 'trac'
templates_path = ['.']
extlinks = {
    'issue': ('https://https://github.com/sisl/mark-koren/AdaptiveStressTestingToolbox/issues/%s', '#'),
    'pr': ('https://https://github.com/sisl/mark-koren/AdaptiveStressTestingToolbox/pull/%s', 'PR #'),
}
linkcheck_retries = 3

# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only set the theme if we're building docs locally
    html_theme = 'sphinx_rtd_theme'

html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
    '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

# sphinx.ext.intersphinx config
intersphinx_mapping = {
    'garage_latest': ('https://garage.readthedocs.io/en/latest/', None),
    'garage': ('https://garage.readthedocs.io/en/v2019.10.1/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}

# sphinx.ext.napoleon config
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



from garage.np.algos import CMAES
from garage.np.baselines import LinearFeatureBaseline
from garage.sampler import OnPolicyVectorizedSampler
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import CategoricalMLPPolicy
from tests.fixtures import snapshot_config, TfGraphTestCase


class TestCMAES(TfGraphTestCase):

    def test_cma_es_cartpole(self):
        """Test CMAES with Cartpole-v1 environment."""
        with LocalTFRunner(snapshot_config) as runner:
            env = TfEnv(env_name='CartPole-v1')

            policy = CategoricalMLPPolicy(name='policy',
                                          env_spec=env.spec,
                                          hidden_sizes=(32, 32))

            baseline = LinearFeatureBaseline(env_spec=env.spec)

            n_samples = 20

            algo = CMAES(env_spec=env.spec,
                         policy=policy,
                         baseline=baseline,
                         max_path_length=100,
                         n_samples=n_samples)

            runner.setup(algo, env, sampler_cls=OnPolicyVectorizedSampler)
            runner.train(n_epochs=1, batch_size=1000, n_epoch_cycles=n_samples)
            # No assertion on return because CMAES is not stable.

            env.close()



'''
This script creates a regression test over garage-DDPG and baselines-DDPG.
It get Mujoco1M benchmarks from baselines benchmark, and test each task in
its trial times on garage model and baselines model. For each task, there will
be `trial` times with different random seeds. For each trial, there will be two
log directories corresponding to baselines and garage. And there will be a plot
plotting the average return curve from baselines and garage.
'''
import datetime
import os
import os.path as osp
import random

from baselines import logger as baselines_logger
from baselines.bench import benchmarks
from baselines.common.misc_util import set_global_seeds
from baselines.ddpg.memory import Memory
from baselines.ddpg.models import Actor, Critic
from baselines.ddpg.noise import OrnsteinUhlenbeckActionNoise
import baselines.ddpg.training as training
from baselines.logger import configure
import dowel
from dowel import logger as dowel_logger
import gym
from mpi4py import MPI
import numpy as np
import pytest
import tensorflow as tf

from garage.envs import normalize
from garage.experiment import deterministic
from garage.np.exploration_strategies import OUStrategy
from garage.replay_buffer import SimpleReplayBuffer
from garage.tf.algos import DDPG
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import ContinuousMLPPolicy
from garage.tf.q_functions import ContinuousMLPQFunction
from tests.fixtures import snapshot_config
import tests.helpers as Rh
from tests.wrappers import AutoStopEnv

# Hyperparams for baselines and garage
params = {
    'policy_lr': 1e-4,
    'qf_lr': 1e-3,
    'policy_hidden_sizes': [64, 64],
    'qf_hidden_sizes': [64, 64],
    'n_epochs': 500,
    'n_epoch_cycles': 20,
    'n_rollout_steps': 100,
    'n_train_steps': 50,
    'discount': 0.9,
    'tau': 1e-2,
    'replay_buffer_size': int(1e6),
    'sigma': 0.2,
}


class TestBenchmarkDDPG:
    '''Compare benchmarks between garage and baselines.'''

    @pytest.mark.huge
    def test_benchmark_ddpg(self):
        '''
        Compare benchmarks between garage and baselines.
        :return:
        '''
        # Load Mujoco1M tasks, you can check other benchmarks here
        # https://github.com/openai/baselines/blob/master/baselines/bench/benchmarks.py
        mujoco1m = benchmarks.get_benchmark('Mujoco1M')

        timestamp = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')
        benchmark_dir = osp.join(os.getcwd(), 'data', 'local', 'benchmarks',
                                 'ddpg', timestamp)
        result_json = {}
        for task in mujoco1m['tasks']:
            env_id = task['env_id']
            env = gym.make(env_id)
            baseline_env = AutoStopEnv(
                env_name=env_id, max_path_length=params['n_rollout_steps'])
            seeds = random.sample(range(100), task['trials'])

            task_dir = osp.join(benchmark_dir, env_id)
            plt_file = osp.join(benchmark_dir,
                                '{}_benchmark.png'.format(env_id))
            relplt_file = osp.join(benchmark_dir,
                                   '{}_benchmark_mean.png'.format(env_id))
            baselines_csvs = []
            garage_csvs = []

            for trial in range(task['trials']):
                env.reset()
                baseline_env.reset()
                seed = seeds[trial]

                trial_dir = osp.join(
                    task_dir, 'trial_{}_seed_{}'.format(trial + 1, seed))
                garage_dir = osp.join(trial_dir, 'garage')
                baselines_dir = osp.join(trial_dir, 'baselines')

                with tf.Graph().as_default():
                    # Run garage algorithms
                    garage_csv = run_garage(env, seed, garage_dir)

                    # Run baselines algorithms
                    baselines_csv = run_baselines(baseline_env, seed,
                                                  baselines_dir)

                garage_csvs.append(garage_csv)
                baselines_csvs.append(baselines_csv)

            env.close()

            Rh.plot(b_csvs=baselines_csvs,
                    g_csvs=garage_csvs,
                    g_x='Epoch',
                    g_y='AverageReturn',
                    g_z='Garage',
                    b_x='total/epochs',
                    b_y='rollout/return',
                    b_z='Baseline',
                    trials=task['trials'],
                    seeds=seeds,
                    plt_file=plt_file,
                    env_id=env_id,
                    x_label='Epoch',
                    y_label='AverageReturn')

            Rh.relplot(g_csvs=garage_csvs,
                       b_csvs=baselines_csvs,
                       g_x='Epoch',
                       g_y='AverageReturn',
                       g_z='Garage',
                       b_x='total/epochs',
                       b_y='rollout/return',
                       b_z='Baseline',
                       trials=task['trials'],
                       seeds=seeds,
                       plt_file=relplt_file,
                       env_id=env_id,
                       x_label='Epoch',
                       y_label='AverageReturn')

            result_json[env_id] = Rh.create_json(
                b_csvs=baselines_csvs,
                g_csvs=garage_csvs,
                seeds=seeds,
                trails=task['trials'],
                g_x='Epoch',
                g_y='AverageReturn',
                b_x='total/epochs',
                b_y='rollout/return',
                factor_g=params['n_epoch_cycles'] * params['n_rollout_steps'],
                factor_b=1)

        Rh.write_file(result_json, 'DDPG')


def run_garage(env, seed, log_dir):
    '''
    Create garage model and training.
    Replace the ddpg with the algorithm you want to run.
    :param env: Environment of the task.
    :param seed: Random seed for the trial.
    :param log_dir: Log dir path.
    :return:
    '''
    deterministic.set_seed(seed)

    with LocalTFRunner(snapshot_config) as runner:
        env = TfEnv(normalize(env))
        # Set up params for ddpg
        action_noise = OUStrategy(env.spec, sigma=params['sigma'])

        policy = ContinuousMLPPolicy(
            env_spec=env.spec,
            hidden_sizes=params['policy_hidden_sizes'],
            hidden_nonlinearity=tf.nn.relu,
            output_nonlinearity=tf.nn.tanh)

        qf = ContinuousMLPQFunction(env_spec=env.spec,
                                    hidden_sizes=params['qf_hidden_sizes'],
                                    hidden_nonlinearity=tf.nn.relu)

        replay_buffer = SimpleReplayBuffer(
            env_spec=env.spec,
            size_in_transitions=params['replay_buffer_size'],
            time_horizon=params['n_rollout_steps'])

        ddpg = DDPG(env_spec=env.spec,
                    policy=policy,
                    qf=qf,
                    replay_buffer=replay_buffer,
                    policy_lr=params['policy_lr'],
                    qf_lr=params['qf_lr'],
                    target_update_tau=params['tau'],
                    n_train_steps=params['n_train_steps'],
                    discount=params['discount'],
                    min_buffer_size=int(1e4),
                    exploration_strategy=action_noise,
                    policy_optimizer=tf.train.AdamOptimizer,
                    qf_optimizer=tf.train.AdamOptimizer)

        # Set up logger since we are not using run_experiment
        tabular_log_file = osp.join(log_dir, 'progress.csv')
        tensorboard_log_dir = osp.join(log_dir)
        dowel_logger.add_output(dowel.StdOutput())
        dowel_logger.add_output(dowel.CsvOutput(tabular_log_file))
        dowel_logger.add_output(dowel.TensorBoardOutput(tensorboard_log_dir))

        runner.setup(ddpg, env)
        runner.train(n_epochs=params['n_epochs'],
                     n_epoch_cycles=params['n_epoch_cycles'],
                     batch_size=params['n_rollout_steps'])

        dowel_logger.remove_all()

        return tabular_log_file


def run_baselines(env, seed, log_dir):
    '''
    Create baselines model and training.
    Replace the ddpg and its training with the algorithm you want to run.
    :param env: Environment of the task.
    :param seed: Random seed for the trial.
    :param log_dir: Log dir path.
    :return
    '''
    rank = MPI.COMM_WORLD.Get_rank()
    seed = seed + 1000000 * rank
    set_global_seeds(seed)
    env.seed(seed)

    # Set up logger for baselines
    configure(dir=log_dir, format_strs=['stdout', 'log', 'csv', 'tensorboard'])
    baselines_logger.info('rank {}: seed={}, logdir={}'.format(
        rank, seed, baselines_logger.get_dir()))

    # Set up params for baselines ddpg
    nb_actions = env.action_space.shape[-1]
    layer_norm = False

    action_noise = OrnsteinUhlenbeckActionNoise(mu=np.zeros(nb_actions),
                                                sigma=float(params['sigma']) *
                                                np.ones(nb_actions))
    memory = Memory(limit=params['replay_buffer_size'],
                    action_shape=env.action_space.shape,
                    observation_shape=env.observation_space.shape)
    critic = Critic(layer_norm=layer_norm)
    actor = Actor(nb_actions, layer_norm=layer_norm)

    training.train(env=env,
                   eval_env=None,
                   param_noise=None,
                   action_noise=action_noise,
                   actor=actor,
                   critic=critic,
                   memory=memory,
                   nb_epochs=params['n_epochs'],
                   nb_epoch_cycles=params['n_epoch_cycles'],
                   render_eval=False,
                   reward_scale=1.,
                   render=False,
                   normalize_returns=False,
                   normalize_observations=False,
                   critic_l2_reg=0,
                   actor_lr=params['policy_lr'],
                   critic_lr=params['qf_lr'],
                   popart=False,
                   gamma=params['discount'],
                   clip_norm=None,
                   nb_train_steps=params['n_train_steps'],
                   nb_rollout_steps=params['n_rollout_steps'],
                   nb_eval_steps=100,
                   batch_size=64)

    return osp.join(log_dir, 'progress.csv')



"""Collects samples in parallel using a stateful pool of workers."""

import tensorflow as tf

from garage.sampler import parallel_sampler
from garage.sampler.base import BaseSampler
from garage.sampler.stateful_pool import singleton_pool
from garage.sampler.utils import truncate_paths


def worker_init_tf(g):
    """Initialize the tf.Session on a worker."""
    g.sess = tf.compat.v1.Session()
    g.sess.__enter__()


def worker_init_tf_vars(g):
    """Initialize the policy parameters on a worker."""
    g.sess.run(tf.compat.v1.global_variables_initializer())


class BatchSampler(BaseSampler):
    """Collects samples in parallel using a stateful pool of workers."""

    def __init__(self, algo, env, n_envs):
        super(BatchSampler, self).__init__(algo, env)
        self.n_envs = n_envs

    def start_worker(self):
        """Initialize the sampler."""
        assert singleton_pool.initialized, (
            'Use singleton_pool.initialize(n_parallel) to setup workers.')
        if singleton_pool.n_parallel > 1:
            singleton_pool.run_each(worker_init_tf)
        parallel_sampler.populate_task(self.env, self.algo.policy)
        if singleton_pool.n_parallel > 1:
            singleton_pool.run_each(worker_init_tf_vars)

    def shutdown_worker(self):
        """Terminate workers if necessary."""
        parallel_sampler.terminate_task(scope=self.algo.scope)

    def obtain_samples(self, itr, batch_size=None, whole_paths=True):
        """Collect samples for the given iteration number."""
        if not batch_size:
            batch_size = self.algo.max_path_length * self.n_envs

        cur_policy_params = self.algo.policy.get_param_values()
        paths = parallel_sampler.sample_paths(
            policy_params=cur_policy_params,
            max_samples=batch_size,
            max_path_length=self.algo.max_path_length,
            scope=self.algo.scope,
        )
        if whole_paths:
            return paths
        else:
            paths_truncated = truncate_paths(paths, batch_size)
            return paths_truncated



"""Class to define the action and observation spaces of an AST problem."""


class ASTSpaces(object):
    """Class to define the action and observation spaces of an AST problem.

    Both the `action_space` and the `observation_space` should be a
    `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_ type.

    The `action_space` is only used to clip actions if `ASTEnv` is wrapped by the normalize env.

    If using `ASTEnv` with `blackbox_sim_state == True`, `observation_space` should define the space for each
    simulation state variable. Otherwise, it should define the space of initial condition variables.

    If using `ASTEnv` with `fixed_init_state == False`, the initial conditions of each rollout will be randomly
    sampled at uniform from the observation_space.

    """

    def __init__(self):
        pass

    @property
    def action_space(self):
        """Returns a definition of the action space of the reinforcement learning problem.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The action space of the reinforcement learning problem.
        """
        raise NotImplementedError

    @property
    def observation_space(self):
        """Returns a definition of the observation space of the reinforcement learning problem.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The observation space of the reinforcement learning problem.
        """
        raise NotImplementedError



"""Samplers which run agents that use Tensorflow in environments."""

from garage.tf.samplers.batch_sampler import BatchSampler
from garage.tf.samplers.ray_sampler import (RaySamplerTF, SamplerWorkerTF)

__all__ = ['BatchSampler', 'RaySamplerTF', 'SamplerWorkerTF']




import gym.utils.seeding as seeding
import numpy as np


class RSG:
    """The random seed generator for AST using random seeds.

    Parameters
    ----------
    state_length : int, optional
        The length of the RSG state.
    seed : int, optional
        The initial seed to generate the initial state.
    """

    def __init__(self, state_length=1, seed=0):
        self.state = seed_to_state_itr(state_length, seed)

    def __eq__(self, other):
        """The redefined equal function.

        Returns
        ----------
        is_equal : bool
            Whether the two RSG are equal.
        """
        return np.array_equal(self.state, other.state)

    def __hash__(self):
        """The redefined hashing function.

        Returns
        ----------
        hash : int
            The hashing result.
        """
        return hash(tuple(self.state))

    def length(self):
        """Return the length of the RSG state.

        Returns
        ----------
        length : int
            The length of the RSG state.
        """
        return len(self.state)

    def next(self):
        """Step the RSG state.
        """
        self.state = np.array(list(map(seeding.hash_seed, self.state)), dtype=np.uint32)

    def set_from_seed(self, length, seed):
        """Set the RSG state using the seed.

        Parameters
        ----------
        length : int
            The length of the RSG state.
        seed : int
            The seed to generate the state.
        """
        self.state = seed_to_state_itr(length, seed)


def seed_to_state_itr(state_length, seed):
    """Generate the RSG state using the seed.

    Parameters
    ----------
    state_length : int
        The length of the RSG state.
    seed : int
        The seed to generate the state.

    Returns
    ----------
    state : :py:class:`numpy.ndarry`
        The generated state.
    """
    state = []
    seedi = seed
    for i in range(state_length):
        # print(seedi)
        state.append(seedi)
        seedi = seeding.hash_seed(seedi)
    return np.array(state, dtype=np.uint32)



"""Optimizers for RL problems"""
# from .direction_constraint_optimizer import PerlmutterHvp
# from .direction_constraint_optimizer import FiniteDifferenceHvp
from .direction_constraint_optimizer import DirectionConstraintOptimizer  # noqa



"""Example simulator wrapper for a scenario of an AV approaching a crosswalk where some pedestrians are crossing."""
import numpy as np  # Used for math

from ast_toolbox.simulators import ASTSimulator  # import parent Simulator class
from ast_toolbox.simulators.example_av_simulator import ToyAVSimulator  # import the simulator to wrap


class ExampleAVSimulator(ASTSimulator):  # Define the class
    """Example simulator wrapper for a scenario of an AV approaching a crosswalk where some pedestrians are crossing.

    Wraps :py:class:`ast_toolbox.simulators.example_av_simulator.ToyAVSimulator`

    Parameters
    ----------
    num_peds : int
        Number of pedestrians crossing the street.
    simulator_args : dict
        Dictionary of keyword arguments to be passed to the wrapped simulator.
    kwargs :
        Keyword arguments passed to the super class.
    """

    def __init__(self,
                 num_peds=1,
                 simulator_args=None,
                 **kwargs):

        # Constant hyper-params -- set by user
        self.c_num_peds = num_peds
        if simulator_args is None:
            simulator_args = {}

        self._action = np.array([0] * (6 * self.c_num_peds))
        self.simulator = ToyAVSimulator(num_peds=num_peds, **simulator_args)

        # initialize the base Simulator
        super().__init__(**kwargs)

    def get_first_action(self):
        """ An initialization method used in Go-Explore.

        Returns
        -------
        array_like
            A 1-D array of the same dimension as the action space, all zeros.
        """
        return np.array([0] * (6 * self.c_num_peds))

    def simulate(self, actions, s_0):
        """Run a full simulation given the AST solver's actions and initial conditions.

        `simulate` takes in the AST solver's actions and the initial conditions. It should return two values: a
        terminal index and an array of relevant simulation information.

        Parameters
        ----------
        actions : list[array_like]
            A sequential list of actions taken by the AST Solver which deterministically control the simulation.
        s_0 : array_like
            An array specifying the initial conditions to set the simulator to.

        Returns
        -------
        terminal_index : int
            The index of the action that resulted in a state in the goal set E. If no state is found
            terminal_index should be returned as -1.
        array_like
            An array of relevant simulator info, which can then be used for analysis or diagnostics.

        """
        return self.simulator.run_simulation(actions=actions, s_0=s_0, simulation_horizon=self.c_max_path_length)

    def closed_loop_step(self, action):
        """User implemented function to step the simulation forward in time when closed-loop control is active.

        This function should step the simulator forward a single timestep based on the given action. It will only
        be called when `open_loop` is False. This function should always return `self.observation_return()`.

        Parameters
        ----------
        action : array_like
            A 1-D array of actions taken by the AST Solver which deterministically control
            a single step forward in the simulation.

        Returns
        -------
        array_like
            An observation from the timestep, determined by the settings and the `observation_return` helper function.
        """
        # grab simulation state, if interactive
        self.observation = np.ndarray.flatten(self.simulator.step_simulation(action))

        return self.observation_return()

    def reset(self, s_0):
        """Resets the state of the environment, returning an initial observation.

        User implementations should always call the super class implementation.
        This function should always return `self.observation_return()`.

        Parameters
        ----------
        s_0 : array_like
            The initial conditions to reset the simulator to.

        Returns
        -------
        array_like
            An observation from the timestep, determined by the settings and the `observation_return` helper function.
        """

        # Call ASTSimulator's reset function (required!)
        super(ExampleAVSimulator, self).reset(s_0=s_0)
        # Reset the simulation
        self.observation = np.ndarray.flatten(self.simulator.reset(s_0))

        return self.observation_return()

    def get_reward_info(self):
        """
        Returns any info needed by the reward function to calculate the current reward.
        """
        # Get the ground truth state from the toy simulator
        sim_state = self.simulator.get_ground_truth()

        return {"peds": sim_state['peds'],
                "car": sim_state['car'],
                "is_goal": self.is_goal(),
                "is_terminal": self.is_terminal()}

    def is_goal(self):
        """
        Returns whether the current state is in the goal set.
        Returns
        -------
        bool
            True if current state is in goal set.
        """
        # Ask the toy simulator if a collision was detected
        return self.simulator.collision_detected()

    def log(self):
        """
        Perform any logging steps.
        """
        # Get the ground truth state from the toy simulator
        sim_state = self.simulator.get_ground_truth()

        # Create a cache of step specific variables for post-simulation analysis
        cache = np.hstack([0.0,  # Dummy, will be filled in with trial # during post processing in save_trials.py
                           sim_state['step'],
                           np.ndarray.flatten(sim_state['car']),
                           np.ndarray.flatten(sim_state['peds']),
                           np.ndarray.flatten(sim_state['action']),
                           np.ndarray.flatten(sim_state['car_obs']),
                           0.0])

        self._info.append(cache)

    def clone_state(self):
        """Clone the simulator state for later resetting.

        This function is used in conjunction with `restore_state` for Go-Explore and Backwards Algorithm
        to do their deterministic resets.

        Returns
        -------
        array_like
            An array of all the simulation state variables.

        """
        # Get the ground truth state from the toy simulator
        simulator_state = self.simulator.get_ground_truth()

        return np.concatenate((np.array([simulator_state['step']]),
                               np.array([simulator_state['path_length']]),
                               np.array([int(simulator_state['is_terminal'])]),
                               simulator_state['car'],
                               simulator_state['car_accel'],
                               simulator_state['peds'].flatten(),
                               simulator_state['car_obs'].flatten(),
                               simulator_state['action'].flatten(),
                               simulator_state['initial_conditions']), axis=0)

    def restore_state(self, in_simulator_state):
        """Reset the simulation deterministically to a previously cloned state.

        This function is used in conjunction with `clone_state` for Go-Explore and Backwards Algorithm
        to do their deterministic resets.

        Parameters
        ----------
        in_simulator_state : array_like
            An array of all the simulation state variables.

        """
        # Put the simulators state variables in dict form
        simulator_state = {}

        simulator_state['step'] = in_simulator_state[0]
        simulator_state['path_length'] = in_simulator_state[1]
        simulator_state['is_terminal'] = bool(in_simulator_state[2])
        simulator_state['car'] = in_simulator_state[3:7]
        simulator_state['car_accel'] = in_simulator_state[7:9]
        peds_end_index = 9 + self.c_num_peds * 4
        simulator_state['peds'] = in_simulator_state[9:peds_end_index].reshape((self.c_num_peds, 4))
        car_obs_end_index = peds_end_index + self.c_num_peds * 4
        simulator_state['car_obs'] = in_simulator_state[peds_end_index:car_obs_end_index].reshape((self.c_num_peds, 4))
        simulator_state['action'] = in_simulator_state[car_obs_end_index:car_obs_end_index + self._action.shape[0]]
        simulator_state['initial_conditions'] = in_simulator_state[car_obs_end_index + self._action.shape[0]:]

        # Set ground truth of actual simulator
        self.simulator.set_ground_truth(simulator_state)

        # Set wrapper state variables
        self._info = []
        self.initial_conditions = np.array(simulator_state['initial_conditions'])
        self._is_terminal = simulator_state['is_terminal']
        self._path_length = simulator_state['path_length']



import ast_toolbox.mcts.MDP as MDP


class AcionSequence:
    """Sturcture storing the actions sequences.

    Parameters
    ----------
    sequence : list
        The list of actions.
    index : int, optional
        The initial action index in the sequence.
    """

    def __init__(self, sequence, index=0):
        self.sequence = sequence
        self.index = index


def action_seq_policy(action_seq, s):
    """The policy wrapper for the action sequence.

    Parameters
    ----------
    action_seq : :py:class:`ast_toolbox.mcts.ASTSim.AcionSequence`
        The action sequence.
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The AST state.

    Returns
    ----------
    action : `ast_toolbox.mcts.AdaptiveStressTesting.ASTAction`
        The AST action.
    """
    action = action_seq.sequence[action_seq.index]
    action_seq.index += 1
    return action


def play_sequence(ast, actions, verbose=False, sleeptime=0.0):
    """Rollout the action sequence.

    Parameters
    ----------
    ast : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.AdaptiveStressTest`
        The AST object.
    actions : list
        The action sequence.
    verbose : bool, optional
        Whether to log the rollout information.
    sleeptime: float, optional
        The pause time between each step.

    Returns
    ----------
    rewards : list[float]
        The rewards.
    actions2 : list
        The action sequence of the path. Should be the same as the input actions.
    """
    rewards, actions2 = MDP.simulate(ast.transition_model, AcionSequence(actions), action_seq_policy, verbose=verbose, sleeptime=sleeptime)
    assert actions == actions2
    return rewards, actions2



import numpy as np
import pytest
import tensorflow as tf

from garage.tf.models.gru import gru
from tests.fixtures import TfGraphTestCase
from tests.helpers import recurrent_step_gru


class TestGRU(TfGraphTestCase):

    def setup_method(self):
        super().setup_method()
        self.batch_size = 2
        self.hidden_dim = 2

        self.step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, self.hidden_dim),
            name='initial_hidden',
            dtype=tf.float32)

        self.gru_cell = tf.keras.layers.GRUCell(
            units=self.hidden_dim,
            activation=tf.nn.tanh,
            kernel_initializer=tf.constant_initializer(1),
            recurrent_activation=tf.nn.sigmoid,
            recurrent_initializer=tf.constant_initializer(1),
            name='lstm_layer')

    # yapf: disable
    @pytest.mark.parametrize('time_step, input_dim, output_dim, '
                             'hidden_init, cell_init', [
        (1, 1, 1, 0, 0),  # noqa: E122
        (1, 1, 3, 0, 0),
        (1, 3, 1, 0, 0),
        (3, 1, 1, 0, 0),
        (3, 3, 1, 0, 0),
        (3, 3, 3, 0, 0),
        (1, 1, 1, 0.5, 0.5),
        (1, 1, 3, 0.5, 0.5),
        (1, 3, 1, 0.5, 0.5),
        (3, 1, 1, 0.5, 0.5),
        (3, 3, 1, 0.5, 0.5),
        (3, 3, 3, 0.5, 0.5),
    ])
    # yapf: enable
    def test_output_shapes(self, time_step, input_dim, output_dim, hidden_init,
                           cell_init):
        obs_inputs = np.full((self.batch_size, time_step, input_dim), 1.)
        obs_input = np.full((self.batch_size, input_dim), 1.)

        input_var = tf.compat.v1.placeholder(tf.float32,
                                             shape=(None, None, input_dim),
                                             name='input')
        step_input_var = tf.compat.v1.placeholder(tf.float32,
                                                  shape=(None, input_dim),
                                                  name='step_input')
        output_nonlinearity = tf.keras.layers.Dense(
            units=output_dim,
            activation=None,
            kernel_initializer=tf.constant_initializer(1))
        with tf.compat.v1.variable_scope('GRU'):
            self.gru = gru(
                all_input_var=input_var,
                name='gru',
                gru_cell=self.gru_cell,
                step_input_var=step_input_var,
                step_hidden_var=self.step_hidden_var,
                hidden_state_init=tf.constant_initializer(hidden_init),
                output_nonlinearity_layer=output_nonlinearity)

        self.sess.run(tf.compat.v1.global_variables_initializer())

        # Compute output by doing t step() on the gru cell
        outputs_t, output_t, h_t, hidden_init = self.gru
        hidden = np.full((self.batch_size, self.hidden_dim),
                         hidden_init.eval())

        for _ in range(time_step):
            output, hidden = self.sess.run([output_t, h_t],
                                           feed_dict={
                                               step_input_var: obs_input,
                                               self.step_hidden_var: hidden,
                                           })  # noqa: E126
            assert output.shape == (self.batch_size, output_dim)
            assert hidden.shape == (self.batch_size, self.hidden_dim)

        full_output = self.sess.run(outputs_t,
                                    feed_dict={input_var: obs_inputs})

        assert full_output.shape == (self.batch_size, time_step, output_dim)

    # yapf: disable
    @pytest.mark.parametrize('time_step, input_dim, output_dim, '
                             'hidden_init, cell_init', [
        (1, 1, 1, 0, 0),  # noqa: E122
        (1, 1, 3, 0, 0),
        (1, 3, 1, 0, 0),
        (3, 1, 1, 0, 0),
        (3, 3, 1, 0, 0),
        (3, 3, 3, 0, 0),
        (1, 1, 1, 0.5, 0.5),
        (1, 1, 3, 0.5, 0.5),
        (1, 3, 1, 0.5, 0.5),
        (3, 1, 1, 0.5, 0.5),
        (3, 3, 1, 0.5, 0.5),
        (3, 3, 3, 0.5, 0.5),
    ])
    # yapf: enable
    def test_output_value(self, time_step, input_dim, output_dim, hidden_init,
                          cell_init):
        obs_inputs = np.full((self.batch_size, time_step, input_dim), 1.)
        obs_input = np.full((self.batch_size, input_dim), 1.)

        input_var = tf.compat.v1.placeholder(tf.float32,
                                             shape=(None, None, input_dim),
                                             name='input')
        step_input_var = tf.compat.v1.placeholder(tf.float32,
                                                  shape=(None, input_dim),
                                                  name='step_input')
        output_nonlinearity = tf.keras.layers.Dense(
            units=output_dim,
            activation=None,
            kernel_initializer=tf.constant_initializer(1))
        with tf.compat.v1.variable_scope('GRU'):
            self.gru = gru(
                all_input_var=input_var,
                name='gru',
                gru_cell=self.gru_cell,
                step_input_var=step_input_var,
                step_hidden_var=self.step_hidden_var,
                hidden_state_init=tf.constant_initializer(hidden_init),
                output_nonlinearity_layer=output_nonlinearity)

        self.sess.run(tf.compat.v1.global_variables_initializer())

        # Compute output by doing t step() on the gru cell
        outputs_t, output_t, h_t, hidden_init = self.gru
        hidden1 = hidden2 = np.full((self.batch_size, self.hidden_dim),
                                    hidden_init.eval())

        for i in range(time_step):
            output1, hidden1 = self.sess.run([output_t, h_t],
                                             feed_dict={
                                                 step_input_var: obs_input,
                                                 self.step_hidden_var: hidden1
                                             })  # noqa: E126

            hidden2 = recurrent_step_gru(input_val=obs_input,
                                         num_units=self.hidden_dim,
                                         step_hidden=hidden2,
                                         w_x_init=1.,
                                         w_h_init=1.,
                                         b_init=0.,
                                         nonlinearity=np.tanh,
                                         gate_nonlinearity=lambda x: 1. /
                                         (1. + np.exp(-x)))

            output_nonlinearity = np.full(
                (np.prod(hidden2.shape[1:]), output_dim), 1.)
            output2 = np.matmul(hidden2, output_nonlinearity)

            assert np.allclose(output1, output2)
            assert np.allclose(hidden1, hidden2)

        full_output1 = self.sess.run(outputs_t,
                                     feed_dict={input_var: obs_inputs})

        hidden2 = np.full((self.batch_size, self.hidden_dim),
                          hidden_init.eval())
        stack_hidden = None
        for i in range(time_step):
            hidden2 = recurrent_step_gru(input_val=obs_inputs[:, i, :],
                                         num_units=self.hidden_dim,
                                         step_hidden=hidden2,
                                         w_x_init=1.,
                                         w_h_init=1.,
                                         b_init=0.,
                                         nonlinearity=np.tanh,
                                         gate_nonlinearity=lambda x: 1. /
                                         (1. + np.exp(-x)))
            if stack_hidden is None:
                stack_hidden = hidden2[:, np.newaxis, :]
            else:
                stack_hidden = np.concatenate(
                    (stack_hidden, hidden2[:, np.newaxis, :]), axis=1)
        output_nonlinearity = np.full((np.prod(hidden2.shape[1:]), output_dim),
                                      1.)
        full_output2 = np.matmul(stack_hidden, output_nonlinearity)
        assert np.allclose(full_output1, full_output2)

    # yapf: disable
    @pytest.mark.parametrize('time_step, input_dim, output_dim', [
        (1, 1, 1),
        (1, 1, 3),
        (1, 3, 1),
        (3, 1, 1),
        (3, 3, 1),
        (3, 3, 3),
    ])
    # yapf: enable
    def test_output_value_trainable_hidden_and_cell(self, time_step, input_dim,
                                                    output_dim):
        obs_inputs = np.full((self.batch_size, time_step, input_dim), 1.)
        obs_input = np.full((self.batch_size, input_dim), 1.)

        input_var = tf.compat.v1.placeholder(tf.float32,
                                             shape=(None, None, input_dim),
                                             name='input')
        step_input_var = tf.compat.v1.placeholder(tf.float32,
                                                  shape=(None, input_dim),
                                                  name='step_input')
        output_nonlinearity = tf.keras.layers.Dense(
            units=output_dim,
            activation=None,
            kernel_initializer=tf.constant_initializer(1))
        with tf.compat.v1.variable_scope('GRU'):
            self.gru = gru(all_input_var=input_var,
                           name='gru',
                           gru_cell=self.gru_cell,
                           step_input_var=step_input_var,
                           step_hidden_var=self.step_hidden_var,
                           hidden_state_init_trainable=True,
                           output_nonlinearity_layer=output_nonlinearity)

        self.sess.run(tf.compat.v1.global_variables_initializer())

        # Compute output by doing t step() on the gru cell
        outputs_t, output_t, h_t, hidden_init = self.gru
        hidden = np.full((self.batch_size, self.hidden_dim),
                         hidden_init.eval())

        _, hidden = self.sess.run([output_t, h_t],
                                  feed_dict={
                                      step_input_var: obs_input,
                                      self.step_hidden_var: hidden,
                                  })  # noqa: E126
        with tf.compat.v1.variable_scope('GRU/gru', reuse=True):
            hidden_init_var = tf.compat.v1.get_variable(name='initial_hidden')
            assert hidden_init_var in tf.compat.v1.trainable_variables()

        full_output1 = self.sess.run(outputs_t,
                                     feed_dict={input_var: obs_inputs})

        hidden2 = np.full((self.batch_size, self.hidden_dim),
                          hidden_init.eval())
        stack_hidden = None
        for i in range(time_step):
            hidden2 = recurrent_step_gru(input_val=obs_inputs[:, i, :],
                                         num_units=self.hidden_dim,
                                         step_hidden=hidden2,
                                         w_x_init=1.,
                                         w_h_init=1.,
                                         b_init=0.,
                                         nonlinearity=np.tanh,
                                         gate_nonlinearity=lambda x: 1. /
                                         (1. + np.exp(-x)))
            if stack_hidden is None:
                stack_hidden = hidden2[:, np.newaxis, :]
            else:
                stack_hidden = np.concatenate(
                    (stack_hidden, hidden2[:, np.newaxis, :]), axis=1)
        output_nonlinearity = np.full((np.prod(hidden2.shape[1:]), output_dim),
                                      1.)
        full_output2 = np.matmul(stack_hidden, output_nonlinearity)
        assert np.allclose(full_output1, full_output2)

    def test_gradient_paths(self):
        time_step = 3
        input_dim = 2
        output_dim = 4
        obs_inputs = np.full((self.batch_size, time_step, input_dim), 1.)
        obs_input = np.full((self.batch_size, input_dim), 1.)

        input_var = tf.compat.v1.placeholder(tf.float32,
                                             shape=(None, None, input_dim),
                                             name='input')
        step_input_var = tf.compat.v1.placeholder(tf.float32,
                                                  shape=(None, input_dim),
                                                  name='step_input')
        output_nonlinearity = tf.keras.layers.Dense(
            units=output_dim,
            activation=None,
            kernel_initializer=tf.constant_initializer(1))
        with tf.compat.v1.variable_scope('GRU'):
            self.gru = gru(all_input_var=input_var,
                           name='gru',
                           gru_cell=self.gru_cell,
                           step_input_var=step_input_var,
                           step_hidden_var=self.step_hidden_var,
                           output_nonlinearity_layer=output_nonlinearity)

        self.sess.run(tf.compat.v1.global_variables_initializer())

        # Compute output by doing t step() on the gru cell
        outputs_t, output_t, h_t, hidden_init = self.gru
        hidden = np.full((self.batch_size, self.hidden_dim),
                         hidden_init.eval())

        grads_step_o_i = tf.gradients(output_t, step_input_var)
        grads_step_o_h = tf.gradients(output_t, self.step_hidden_var)
        grads_step_h = tf.gradients(h_t, step_input_var)

        self.sess.run([grads_step_o_i, grads_step_o_h, grads_step_h],
                      feed_dict={
                          step_input_var: obs_input,
                          self.step_hidden_var: hidden,
                      })  # noqa: E126

        grads_full = tf.gradients(outputs_t, input_var)
        self.sess.run(grads_full, feed_dict={input_var: obs_inputs})

        grads_step_o_i = tf.gradients(outputs_t, step_input_var)
        grads_step_o_h = tf.gradients(outputs_t, self.step_hidden_var)
        grads_step_h = tf.gradients(h_t, input_var)

        # No gradient flow
        with pytest.raises(TypeError):
            self.sess.run(grads_step_o_i,
                          feed_dict={
                              step_input_var: obs_input,
                              self.step_hidden_var: hidden,
                          })
        with pytest.raises(TypeError):
            self.sess.run(grads_step_o_h,
                          feed_dict={
                              step_input_var: obs_input,
                              self.step_hidden_var: hidden,
                          })
        with pytest.raises(TypeError):
            self.sess.run(grads_step_h, feed_dict={input_var: obs_inputs})

    # yapf: disable
    @pytest.mark.parametrize('time_step, input_dim, output_dim, '
                             'hidden_init, cell_init', [
        (1, 1, 1, 0, 0),  # noqa: E122
        (1, 1, 3, 0, 0),
        (1, 3, 1, 0, 0),
        (3, 1, 1, 0, 0),
        (3, 3, 1, 0, 0),
        (3, 3, 3, 0, 0),
        (1, 1, 1, 0.5, 0.5),
        (1, 1, 3, 0.5, 0.5),
        (1, 3, 1, 0.5, 0.5),
        (3, 1, 1, 0.5, 0.5),
        (3, 3, 1, 0.5, 0.5),
        (3, 3, 3, 0.5, 0.5),
    ])
    # yapf: enable
    def test_output_same_as_rnn(self, time_step, input_dim, output_dim,
                                hidden_init, cell_init):
        obs_inputs = np.full((self.batch_size, time_step, input_dim), 1.)
        obs_input = np.full((self.batch_size, input_dim), 1.)

        input_var = tf.compat.v1.placeholder(tf.float32,
                                             shape=(None, None, input_dim),
                                             name='input')
        step_input_var = tf.compat.v1.placeholder(tf.float32,
                                                  shape=(None, input_dim),
                                                  name='step_input')
        output_nonlinearity = tf.keras.layers.Dense(
            units=output_dim,
            activation=None,
            kernel_initializer=tf.constant_initializer(1))
        with tf.compat.v1.variable_scope('GRU'):
            self.gru = gru(
                all_input_var=input_var,
                name='gru',
                gru_cell=self.gru_cell,
                step_input_var=step_input_var,
                step_hidden_var=self.step_hidden_var,
                hidden_state_init=tf.constant_initializer(hidden_init),
                output_nonlinearity_layer=output_nonlinearity)

        self.sess.run(tf.compat.v1.global_variables_initializer())

        # Create a RNN and compute the entire outputs
        rnn_layer = tf.keras.layers.RNN(cell=self.gru_cell,
                                        return_sequences=True,
                                        return_state=True)

        # Set initial state to all 0s
        hidden_var = tf.compat.v1.get_variable(
            name='initial_hidden',
            shape=(self.batch_size, self.hidden_dim),
            initializer=tf.constant_initializer(hidden_init),
            trainable=False,
            dtype=tf.float32)
        outputs, hiddens = rnn_layer(input_var, initial_state=[hidden_var])
        outputs = output_nonlinearity(outputs)

        self.sess.run(tf.compat.v1.global_variables_initializer())

        outputs, hiddens = self.sess.run([outputs, hiddens],
                                         feed_dict={input_var: obs_inputs})

        # Compute output by doing t step() on the gru cell
        hidden = np.full((self.batch_size, self.hidden_dim), hidden_init)
        _, output_t, hidden_t, _ = self.gru
        for i in range(time_step):
            output, hidden = self.sess.run([output_t, hidden_t],
                                           feed_dict={
                                               step_input_var: obs_input,
                                               self.step_hidden_var: hidden,
                                           })  # noqa: E126
            # The output from i-th timestep
            assert np.array_equal(output, outputs[:, i, :])
        assert np.array_equal(hidden, hiddens)

        # Also the full output from lstm
        full_outputs = self.sess.run(self.gru[0],
                                     feed_dict={input_var: obs_inputs})
        assert np.array_equal(outputs, full_outputs)



"""GaussianMLPPolicy."""
import torch

from garage.torch.modules import GaussianMLPModule
from garage.torch.policies import Policy


class GaussianMLPPolicy(Policy, GaussianMLPModule):
    """
    GaussianMLPPolicy.

    A policy that contains a MLP to make prediction based on a gaussian
    distribution.

    Args:
        env_spec (garage.envs.env_spec.EnvSpec): Environment specification.
        module : GaussianMLPModule to make prediction based on a gaussian
        distribution.
    :return:
    """

    def __init__(self, env_spec, **kwargs):
        self._obs_dim = env_spec.observation_space.flat_dim
        self._action_dim = env_spec.action_space.flat_dim

        Policy.__init__(self, env_spec)
        GaussianMLPModule.__init__(self,
                                   input_dim=self._obs_dim,
                                   output_dim=self._action_dim,
                                   **kwargs)

    def forward(self, inputs):
        """Forward method."""
        return super().forward(torch.Tensor(inputs))

    def get_action(self, observation):
        """Get a single action given an observation."""
        with torch.no_grad():
            observation = observation.unsqueeze(0)
            dist = self.forward(observation)
            return (dist.rsample().squeeze(0).numpy(),
                    dict(mean=dist.mean.squeeze(0).numpy(),
                         log_std=(dist.variance**.5).log().squeeze(0).numpy()))

    def get_actions(self, observations):
        """Get actions given observations."""
        with torch.no_grad():
            dist = self.forward(observations)
            return (dist.rsample().numpy(),
                    dict(mean=dist.mean.numpy(),
                         log_std=(dist.variance**.5).log().numpy()))

    def log_likelihood(self, observation, action):
        """Get log likelihood given observations and action."""
        dist = self.forward(observation)
        return dist.log_prob(action)

    def get_entropy(self, observation):
        """Get entropy given observations."""
        dist = self.forward(observation)
        return dist.entropy()

    def reset(self, dones=None):
        """Reset the environment."""
        pass

    @property
    def vectorized(self):
        """Vectorized or not."""
        return True



"""gym.Env wrappers.

Used to transform an environment in a modular way.
It is also possible to apply multiple wrappers at the same
time.

Example:
    StackFrames(GrayScale(gym.make('env')))

"""
from garage.envs.wrappers.atari_env import AtariEnv
from garage.envs.wrappers.clip_reward import ClipReward
from garage.envs.wrappers.episodic_life import EpisodicLife
from garage.envs.wrappers.fire_reset import FireReset
from garage.envs.wrappers.grayscale import Grayscale
from garage.envs.wrappers.max_and_skip import MaxAndSkip
from garage.envs.wrappers.noop import Noop
from garage.envs.wrappers.resize import Resize
from garage.envs.wrappers.stack_frames import StackFrames

__all__ = [
    'AtariEnv', 'ClipReward', 'EpisodicLife', 'FireReset', 'Grayscale',
    'MaxAndSkip', 'Noop', 'Resize', 'StackFrames'
]
